/*
 * Carlos engine rewrite
 * Copyright (C) 2006-2011 Gregory Montoir (cyx@users.sourceforge.net)
 */

#ifndef COMMON_H__
#define COMMON_H__

#include "util.h"

typedef struct anim_data_t {
	int8 unk0;
	int16 lifes;
	int8 cycles;
	int8 unk3;
	uint8 unk4;
	uint8 unk5;
	uint8 unk6;
	uint8 init_sprite_num;
	uint8 colliding_opcode;
	uint8 logic_opcode;
	int8 sound_num;
	uint8 rnd;
	uint16 sprite_num;
	uint16 sprite_flags;
	uint16 default_sprite_num;
	uint16 default_sprite_flags;
	uint16 anim_w;
	uint16 anim_h;
	uint16 score;
	int16 bounding_box_x1;
	int16 bounding_box_x2;
	int16 bounding_box_y1;
	int16 bounding_box_y2;
	const uint8 *move_data_ptr;
	int16 distance_dx;
	int16 distance_dy;
	struct anim_data_t *anim_data1_ptr;
	struct anim_data_t *anim_data2_ptr;
	struct anim_data_t *anim_data3_ptr;
	int16 dx;
	int16 dy;
	struct anim_data_t *anim_data4_ptr;
	int16 dx2;
	int16 dy2;
	int16 draw_dx;
	int16 draw_dy;
} anim_data_t;

struct object_state_t {
	uint8 displayed;
	int16 life;
	int8 anim_data_cycles;
	int8 unk3;
	int16 map_pos_x; /* init_map_pos_x ? */
	int16 map_pos_y; /* init_map_pos_y ? */
	anim_data_t *anim_data1_ptr;
	anim_data_t *anim_data2_ptr;
	anim_data_t *anim_data3_ptr;
	anim_data_t *anim_data4_ptr;
	const uint8 *move_data_ptr;
	int16 distance_pos_x;
	int16 distance_pos_y;
	int16 pos_x;
	int16 pos_y;
	uint16 ref_ref_index;
	uint8 unk26;
	uint8 unk27;
	uint8 transparent_flag;
	uint8 unk29;
	uint8 tile_num;
	uint8 visible;
int16 anim_data3_ptr_, anim_data4_ptr_, anim_data4_seg_;
	int16 unk2C;
	int16 unk2E;

	void init_from_anim_data(anim_data_t *ad) {
		anim_data1_ptr = ad;
		anim_data3_ptr = ad;
		anim_data4_ptr = ad->anim_data2_ptr;
		distance_pos_x = ad->distance_dx;
		distance_pos_y = ad->distance_dy;
		move_data_ptr = ad->move_data_ptr;
		anim_data_cycles = ad->cycles;
		unk3 = ad->unk3;
		unk26 = ad->unk4;
		unk27 = ad->unk5;
	}
	void set_to_initial_pos(anim_data_t *ad) {
		int x, y;
		if (ad->move_data_ptr) {
			if (move_data_ptr[0] == 0x89) {
assert(0);
			} else {
				if (move_data_ptr[0] == 0x88) {
					move_data_ptr = ad->move_data_ptr;
				}
				x = (int8)move_data_ptr[0];
				y = (int8)move_data_ptr[1];
				move_data_ptr += 2;
				pos_x += x;
				pos_y += y;
			}
		} else {
			x = distance_pos_x;
			y = distance_pos_y;
			if (unk2C != 0) {
				unk2C = 0;
				x = unk2E;
			}
			pos_x += x;
			pos_y += y;
		}
	}
	anim_data_t *init_object_from_anim3(anim_data_t *ad) {
		object_state_t *os = this;
		anim_data_t *next = ad->anim_data3_ptr;
		os->pos_x += ad->dx;
		os->pos_y += ad->dy;
		os->anim_data_cycles = next->cycles;
		os->unk3 = ad->unk3;
		os->unk26 = next->unk4;
		os->unk27 = next->unk5;
		os->life = next->lifes;
		os->distance_pos_x = next->distance_dx;
		os->move_data_ptr = next->move_data_ptr;
		os->anim_data4_ptr = next->anim_data2_ptr;
		os->distance_pos_y = next->distance_dy;
		os->anim_data1_ptr = next;
		os->anim_data3_ptr = next;
		return next;
	}
};

struct anim_frame_t {
	uint8 frames_count;
	uint8 action_op3_flag;
	int16 frame_w;
	int16 frame_h;
	int16 _null1;
	int16 frame_w2;
	int16 delta_y;
	int16 frame_h2;
	int16 sprite_num;
	const anim_frame_t *next_anim_frame;
	int8 draw_dy;
	int8 draw_dx;
};

typedef struct {
	int16 delta_x;
	int16 delta_y;
	int16 tile_map_offs_x;
	int16 tile_map_offs_y;
	int16 tile_map_end_offs_x;
	int16 tile_map_end_offs_y;
	int16 scroll_dx;
	int16 scroll_dy;
	int16 tile_block_x;
	int16 tile_block_y;
	int16 offs_x;
	int16 offs_y;
	int16 scroll_inc;
} decor_state_t;

typedef struct {
	uint16 tile_num;
	uint16 anim;
	uint16 tiles_table[4];
} tile_anim_t;

typedef struct {
	int score;
	int extra_life_score;
	uint8 boss_fight_counter;
	uint8 boss_explosion_counter;
	int8 tile_cycling_counter;
	tile_anim_t tile_anims_table[6];
} game_state_t;

#define MODE_WALKING 0
#define MODE_CLIMBING 1
#define MODE_JUMPING_FALLING 2

#define HDIR_FLAG_FACING_RIGHT   (1 << 0)
#define HDIR_FLAG_FALLING        (1 << 2)
#define HDIR_FLAG_FACING_CHANGED (1 << 3)
#define HDIR_FLAG_ON_STAIRS_UP   (1 << 4)
#define HDIR_FLAG_ON_STAIRS_DOWN (1 << 5)
#define HDIR_FLAG_CROUCHED       (1 << 6)

#define BIRD_DIR_FLAG_GOING_DOWN (1 << 2)

#define DECOR_REF_FLAG0_BLOCK_DOWN  (1 << 0)
#define DECOR_REF_FLAG0_BLOCK_UP    (1 << 1)
#define DECOR_REF_FLAG0_BLOCK_RIGHT (1 << 2)
#define DECOR_REF_FLAG0_BLOCK_LEFT  (1 << 3)

typedef struct {
	uint8 unk0;
	uint8 unk1; /* MODE_* */
	uint8 hdir_flags; /* HDIR_FLAG_* */
	uint8 vdir_flags;
	int16 pos_x;
	int16 pos_y;
	int16 dim_w;
	int16 dim_h;
	int16 pos_dx_1;
	int16 pos_dy_1;
	int16 pos_dx_2;
	int16 pos_dy_2;
	const uint16 *move_offsets_data_ptr;
	int16 anim_frames_count;
	uint8 fire_button_counter;
	uint8 throwing_ball_counter;
	uint8 decor_ref_flag0; /* DECOR_REF_FLAG0_* */
	uint8 fill_energy_bar;
	uint8 lifes_count;
	uint8 keys_count;
	uint8 rams_count;
	uint8 winds_count;
	uint8 decor_ref_flag4; /* collision with decor */
	uint8 decor_ref_flag3; /* interaction with decor (stairs etc.) */
	uint8 dead_flag;
	uint8 decor_ref_flag2; /* play sound */
	uint8 has_blue_ball;
	int8 throwing_apple_counter;
	uint8 energy;
	uint8 colliding;
	uint8 action_code;
	int16 ram_duration;
	int16 has_wood;
	int16 shield_duration;
	int16 tile_blinking_duration;
	int16 has_spring;
	int16 tilemap_offset;
	uint16 next_tilemap_offset;
	const anim_frame_t *anim_frame_ptr;
	uint8 colliding_flag;
	uint8 rock_hit_counter; /* Nicky2 specific */
	uint8 bird_move_speed; /* Nicky2 specific */
	uint8 bird_dir_flags; /* Nicky2 specific */
	uint8 bird_counter; /* Nicky2 specific */
	const uint16 *bird_move_dy_ptr;
	int16 unk3D; /* Carlos specific */
	uint8 power;
	uint8 increment_power_bar;
} player_state_t;

struct oscar_state_t {
	uint8 unk0;
	uint8 hdir_flags;
	int16 pos_x;
	int16 pos_y;
	int16 dx;
	int16 dy;
	int16 frame_counter;
	int16 frame_num;

	void resetDirection() {
		hdir_flags |= 2;
		dx = dy = 0;
		frame_counter = 0;
		frame_num = 0;
	}
	void resetPosition(int x, int y, int16 *posTable) {
		unk0 = 0xFF;
		hdir_flags = 0;
		pos_y = y;
		pos_x = x;
		frame_num = 0;
		frame_counter = 0;
		dx = 0;
		dy = 0;
		for (int i = 0, pos = 0; i < 20; ++i) {
			posTable[pos++] = pos_x;
			posTable[pos++] = pos_y;
		}
	}
};

struct boss_state_t {
	int16 pos_y;
	int16 pos_x;
	uint16 frame_counter;
	uint16 frame_num;
	uint16 last_frame_num;
	uint16 frame_counter_mask;
};

#endif // COMMON_H__
