/*
 * Carlos engine rewrite
 * Copyright (C) 2006-2011 Gregory Montoir (cyx@users.sourceforge.net)
 */

#include <math.h>
#include <sys/time.h>
#ifdef _WIN32
#include <windows.h>
#define DYNLIB_SYMBOL DECLSPEC_EXPORT
#else
#define DYNLIB_SYMBOL
#endif
#define VARDECL
#include "game.h"
#include "fileio.h"
#include "globals.h"
#include "input.h"
#include "render.h"
#include "sound.h"
#include "stub.h"

static void game_draw_bonuses() {
	if (g_game.bonus_count == 0) {
		return;
	}
	g_game.bonus_count = 10;
	for (int i = 0; i < 10; ++i) {
		object_state_t *os = &g_game.objects_table_ptr8[i];
		if (os->displayed == 0) {
			--g_game.bonus_count;
			continue;
		}
		anim_data_t *ad = os->anim_data1_ptr;
		if (ad->logic_opcode == 2) {
			game_logic_op2(os, ad);
			if (os->displayed == 0) {
				--g_game.bonus_count;
				continue;
			}
		} else {
			if (os->unk27 != 0 && --os->unk27 == 0) {
				if (os->anim_data4_ptr == 0) {
					goto hide_object;
				}
				ad = os->anim_data4_ptr;
				os->init_from_anim_data(ad);
			} else if (ad->cycles != 0 && --os->anim_data_cycles == 0) {
				if (ad->anim_data1_ptr == 0) {
					goto hide_object;
				} else if (ad->anim_data1_ptr == &g_game._res._ref_ref_use_ptr3) {
					if (ad->anim_data3_ptr == 0) {
						goto hide_object;
					}
					ad = ad->anim_data3_ptr;
				} else {
					ad = ad->anim_data1_ptr;
				}
				os->anim_data1_ptr = ad;
				os->anim_data_cycles = ad->cycles;
			}
			os->set_to_initial_pos(ad);
		}
		if (os->pos_x >= -128 && os->pos_x <= GAME_SCREEN_W + 96 && os->pos_y >= -128 && os->pos_y <= GAME_SCREEN_H + 64) {
			game_draw_object(os);
			continue;
		}
hide_object:
		os->displayed = 0;
		--g_game.bonus_count;
	}
}

void Game::init() {
	g_game._render = 0;
	g_game._res.init_static_data__v3();
	objects_table_ptr1 = &objects_table[0];
	objects_table_ptr2 = &objects_table[342];
	objects_table_ptr7 = &objects_table[340];
	objects_table_ptr8 = &objects_table[330];
	objects_table_ptr9 = &objects_table[335];
	setRandomSeed();
}

void game_init_level(int level) {
	g_game.init_player_state();
	g_game._currentLevel = level;
	debug(DBG_GAME, "game_init_level() level=%d", g_game._currentLevel);
//	screen_fade_in_pal
//	res_load_image(20 + _currentLevel);
//	screen_set_pal(LEVELS_PAL + 32 * _currentLevel);
	g_game.op_colliding_mask = 0;
	g_game._res.set_colors(level, g_game._render);
	g_game._res.load_level_data__v3(level, g_game._render);
//	g_game._res.load_level_spr(level);
	game_init_carlos();
	inp_start_level(g_game._currentLevel);
	g_game.execute_op_logic_helper1_flag = 0;
	g_game.new_level_flag = 0;
	_snd_current_sfx_num = -1;
	_snd_current_sfx_priority = 0;
	g_game.player_state.tilemap_offset = 0;
	g_game.player_state.next_tilemap_offset = 0;
	g_game.player_state.has_spring = 0;
	g_game.carlos_move_offsets_ptr = (const uint16 *)&move_data[1];
	g_game.monster_fight_flag = 0;
	g_game.quit_level_flag = 0;
//	byte_18456 = 0;
	g_game.game_state.boss_fight_counter = 0;
	g_game.player_state.winds_count = 0;
	g_game.player_state.tilemap_offset = _screen_cdg_tile_map_start_x * _screen_cdg_tile_map_h + _screen_cdg_tile_map_start_y - 1;
	g_game.fix_player_pos_tilemap_offset();
	g_game.ref_ref_1 = &g_game._res._ref_ref[1];
	g_game.ref_ref_21 = &g_game._res._ref_ref[21];
	game_init_ref_ref(g_game._res._ref_ref, 0x8000);
	game_init_objects_from_positref(g_game._res._posit_ref, 0);
	game_init_level_start_screen();
	g_game.cycle_score = 0;
	g_game.update_nicky_anim_flag = 1;
}

void game_init_carlos() {
	debug(DBG_GAME, "game.init_carlos() pos %d,%d", g_game.player_pos_x, g_game.player_pos_y );
	g_game.player_state.pos_x = g_game.player_pos_x;
	g_game.player_state.pos_y = g_game.player_pos_y;
	g_game.player_state.unk1 = 0;
	g_game.player_state.hdir_flags = 1;
	g_game.player_state.vdir_flags = 0;
	g_game.player_state.anim_frames_count = 0;
	g_game.player_state.throwing_ball_counter = 0;
	g_game.player_state.fill_energy_bar = 1;
	g_game.player_state.dead_flag = 0;
	g_game.player_state.energy = 0;
	g_game.player_state.colliding = 0;
	g_game.player_state.unk3D = 0;
	g_game.player_state.anim_frame_ptr = &carlos_anim_frames_table__v3[19];
	g_game.player_state.shield_duration = 25;
	if (g_game.monster_fight_flag) {
		g_game.player_state.shield_duration >>= 1;
	}
	if (g_game.player_state.winds_count == 0) {
		g_game.player_state.winds_count = 1;
	}
	if (g_game.player_state.rams_count == 0) {
		g_game.player_state.rams_count = 1;
	}
	if (g_game.player_state.has_spring == 0) {
		g_game.player_state.has_spring = 93;
	}
	g_game.unk_ref_index = 0;
	g_game.life_bar_flag = 0;
	g_game.cycle_counter = 0;
	g_game.boss_colliding_op_flag = 0;
	g_game.carlos_direction = 0;
	g_game.carlos_offset_dy = 0;
}

static void game_translate_object_table(object_state_t *os, int num, int16 dx, int16 dy) {
	for (int i = 0; i < num; ++i, ++os) {
		os->pos_x += dx;
		os->pos_y += dy;
	}
}

static void game_translate_object_tables(int16 dx, int16 dy) {
	game_translate_object_table(g_game.objects_table_ptr1, g_game._res._posit_count, dx, dy);
	game_translate_object_table(g_game.objects_table_ptr8, 10, dx, dy);
	game_translate_object_table(g_game.objects_table_ptr2, 7, dx, dy);
//	g_game.execute_op_logic_helper1_flag = -1;
}

static void game_adjust_objects_position_hscroll() {
	static const int scroll_x1 = (GAME_SCREEN_W - GAME_SCROLL_W) / 2;
	static const int scroll_x2 = GAME_SCREEN_W - (GAME_SCREEN_W - GAME_SCROLL_W) / 2;
	int x_min, dx, tile_dx;

	if (g_game.player_state.pos_x <= scroll_x1) { /* x - 16 <= 132 */
		if (g_game.monster_fight_flag) {
			/* set scrolling to 320 during boss sequences (monster 'logic' assumes that screen width) */
			x_min = (_screen_cdg_tile_map_w - MAX(320, GAME_SCREEN_W) / 16) * _screen_cdg_tile_map_h;
		} else {
			x_min = 0;
		}
		if (g_game.decor_state.tile_map_offs_x <= x_min) {
			g_game.decor_state.tile_map_offs_x = x_min;
			g_game.decor_state.scroll_dx = 0;
			return;
		}
		dx = scroll_x1 - g_game.player_state.pos_x;
		if (g_game.decor_state.tile_map_offs_x <= x_min + _screen_cdg_tile_map_h && g_game.decor_state.scroll_dx + dx > 16) {
			dx = 16 - g_game.decor_state.scroll_dx;
		}
		game_translate_object_tables(dx, 0);
		g_game.player_state.pos_x += dx;
		g_game.decor_state.scroll_dx += dx;
		if (g_game.decor_state.scroll_dx > 0) {
			tile_dx = g_game.decor_state.scroll_dx / 16;
			g_game.decor_state.tile_map_offs_x -= tile_dx * _screen_cdg_tile_map_h;
		}
	} else if (g_game.player_state.pos_x + 16 >= scroll_x2) { /* x >= 188 */
		dx = g_game.player_state.pos_x + 16 - scroll_x2;
		if (g_game.decor_state.tile_map_offs_x >= g_game.decor_state.tile_map_end_offs_x - _screen_cdg_tile_map_h) {
			g_game.decor_state.tile_map_offs_x = g_game.decor_state.tile_map_end_offs_x - _screen_cdg_tile_map_h;
			if (g_game.decor_state.scroll_dx < dx) {
				dx = g_game.decor_state.scroll_dx;
			}
		}
		game_translate_object_tables(-dx, 0);
		g_game.player_state.pos_x -= dx;
		g_game.decor_state.scroll_dx -= dx;
		if (g_game.decor_state.scroll_dx < 0) {
			tile_dx = (15 - g_game.decor_state.scroll_dx) / 16;
			g_game.decor_state.tile_map_offs_x += tile_dx * _screen_cdg_tile_map_h;
		}
	}
}

static void game_adjust_objects_position_vscroll() {
	static const int scroll_y1 = 4 * 16;
	static const int scroll_y2 = GAME_SCREEN_H - 4 * 16 - 9;
/*	static const int scroll_y2 = 5 * 16 + 7;*/
	int dy, tile_dy;

	if (g_game.monster_fight_flag) {
		return;
	}

	if (g_game.player_state.pos_y <= scroll_y1) { /* 64 */
		if (g_game.decor_state.tile_map_offs_y <= 0) {
			g_game.decor_state.tile_map_offs_y = 0;
			g_game.decor_state.scroll_dy = 0;
			return;
		}
		dy = scroll_y1 - g_game.player_state.pos_y;
		if (dy > 15) {
			dy = 15;
		}
		if (g_game.decor_state.tile_map_offs_y <= 1) {
			g_game.decor_state.tile_map_offs_y = 1;
			if (g_game.decor_state.scroll_dy + dy > 16) {
				dy = 16 - g_game.decor_state.scroll_dy;
			}
		}
		game_translate_object_tables(0, dy);
		g_game.player_state.pos_y += dy;
		g_game.decor_state.scroll_dy += dy;
		if (g_game.decor_state.scroll_dy > 0) {
			tile_dy = g_game.decor_state.scroll_dy / 16;
			g_game.decor_state.tile_map_offs_y -= tile_dy;
		}
	} else if (g_game.player_state.pos_y >= scroll_y2) { /* 87 */
		dy = g_game.player_state.pos_y - scroll_y2;
		if (dy > 15) {
			dy = 15;
		}
		if (g_game.decor_state.tile_map_offs_y >= g_game.decor_state.tile_map_end_offs_y) {
			g_game.decor_state.tile_map_offs_y = g_game.decor_state.tile_map_end_offs_y;
			if (g_game.decor_state.scroll_dy < dy) {
				dy = g_game.decor_state.scroll_dy;
			}
		}
		game_translate_object_tables(0, -dy);
		g_game.player_state.pos_y -= dy;
		g_game.decor_state.scroll_dy -= dy;
		if (g_game.decor_state.scroll_dy < 0) {
			tile_dy = (15 - g_game.decor_state.scroll_dy) / 16;
			g_game.decor_state.tile_map_offs_y += tile_dy;
		}
	}
}

static void game_adjust_objects_position() {
	if (g_game.player_state.dead_flag == 0) {
		game_adjust_objects_position_hscroll();
		g_game.decor_state.scroll_dx &= 15;
		g_game.player_state.pos_dx_2 = -g_game.decor_state.scroll_dx;
		game_adjust_objects_position_vscroll();
		g_game.decor_state.scroll_dy &= 15;
		g_game.player_state.pos_dy_2 = -g_game.decor_state.scroll_dy;
	}

	g_game.decor_state.offs_x = g_game.player_state.pos_dx_2;
	g_game.decor_state.offs_y = g_game.player_state.pos_dy_2;
	g_game.decor_state.tile_block_x = g_game.decor_state.tile_map_offs_x;
	g_game.decor_state.tile_block_y = g_game.decor_state.tile_map_offs_y;
}

static void game_move_object(object_state_t *os) {
	os->distance_pos_x = 0;
	os->distance_pos_y = 0;
	const int dy = g_game.player_state.pos_y + 28 + g_game.player_state.anim_frame_ptr->delta_y - os->pos_y;
	os->map_pos_y = (dy < 0) ? -dy : dy;
	const int dx = g_game.player_state.pos_x + 24 - os->pos_x;
	os->map_pos_x = (dx < 0) ? -dx : dx;
	os->anim_data3_ptr_ = (os->map_pos_x > os->map_pos_y) ? os->map_pos_x : os->map_pos_y;
	if (dx != 0) {
		if (dx < 0) {
			--os->distance_pos_x;
		} else {
			++os->distance_pos_x;
		}
	}
	if (dy != 0) {
		if (dy < 0) {
			--os->distance_pos_y;
		} else {
			++os->distance_pos_y;
		}
	}
	os->anim_data4_ptr_ = 0;
	os->anim_data4_seg_ = 0;
}

static void game_update_objects2() {
	for (int i = 0; i < 7; ++i) {
		object_state_t *os = &g_game.objects_table_ptr2[i];
		if (os->displayed) {
			anim_data_t *ad = os->anim_data1_ptr;
			if (ad->cycles != 0 && --os->anim_data_cycles == 0) {
				if (ad->anim_data1_ptr == 0) {
					os->displayed = 0;
					continue;
				}
				ad = ad->anim_data1_ptr;
				os->anim_data1_ptr = ad;
				os->anim_data_cycles = ad->cycles;
				os->unk26 = ad->unk4;
			}
			if (os->unk26 & 0x10) {
				if (ad->move_data_ptr != 0) {
					if (os->move_data_ptr[0] == 0x88) {
						os->move_data_ptr = ad->move_data_ptr;
					}
					os->pos_x += (int8)os->move_data_ptr[0];
					os->pos_y += (int8)os->move_data_ptr[1];
					os->move_data_ptr += 2;
				} else {
					os->pos_x += ad->distance_dx;
					os->pos_y += ad->distance_dy;
				}
			} else {
				int j, count = 3;
				int16 delta = os->anim_data3_ptr_;
				if (g_game.monster_fight_flag) {
					if (g_game._currentLevel > 2) {
						count = g_game._currentLevel + 1;
					} else if (g_game._currentLevel > 1 && g_game.cycle_counter != 0) {
						count = 4;
					}
				}
				for (j = 0; j < count; ++j) {
					os->anim_data4_seg_ += os->map_pos_y;
					os->anim_data4_ptr_ += os->map_pos_x;
					if (delta < os->anim_data4_ptr_) {
						os->anim_data4_ptr_ -= delta;
						os->pos_x += os->distance_pos_x;
					}
					if (delta < os->anim_data4_seg_) {
						os->anim_data4_seg_ -= delta;
						os->pos_y += os->distance_pos_y;
					}
				}
			}
			os->pos_x -= g_game.decor_state.delta_x;
			if (os->pos_x >= -8 && os->pos_x <= GAME_SCREEN_W) {
				os->pos_y -= g_game.decor_state.delta_y;
				if (os->pos_y >= -8 && os->pos_y <= GAME_SCREEN_H) {
					game_draw_object(os);
					continue;
				}
			}
			os->displayed = 0;
		}
	}
}

static void game_update_tile_offsets() {
	--g_game.game_state.tile_cycling_counter;
	if (g_game.game_state.tile_cycling_counter < 0) {
		g_game.game_state.tile_cycling_counter = 3;
		for (int i = 0; i < 6; ++i) {
			tile_anim_t *ta = &g_game.game_state.tile_anims_table[i];
			++ta->anim;
			if (ta->anim == 3) {
				ta->anim = 0;
			}
			_screen_cdg_offsets[ta->tile_num] = ta->tiles_table[ta->anim];
		}
	}
}

void game_update_life_bar() {
	if ((g_game.nicky_anim_flags1 & 0xFF) != 0) {
		g_game.player_state.colliding = 0;
	}
	if (g_game.player_state.fill_energy_bar) {
		if (g_game.player_state.energy >= 100) {
			g_game.player_state.fill_energy_bar = 0;
			g_game.life_bar_flag = 0;
			return;
		}
		++g_game.player_state.energy;
		if (g_game.player_state.fill_energy_bar != 1) {
			--g_game.life_bar_flag;
			if (g_game.life_bar_flag == 0) {
				g_game.player_state.fill_energy_bar = 0;
			}
		}
		if (g_game.player_state.pos_y >= GAME_SCREEN_H) {
			g_game.player_state.pos_y = GAME_SCREEN_H + 2;
		}
		return;
	}
	if (g_game.player_state.pos_y >= GAME_SCREEN_H) {
		g_game.player_state.pos_y = GAME_SCREEN_H + 2;
	} else {
		if (g_game.player_state.shield_duration > 0) {
			return;
		}
		if (g_game.update_nicky_anim_flag == 0) {
			return;
		}
		if (g_game.player_state.colliding == 0) {
			return;
		}
	}
	debug(DBG_GAME, "game_update_life_bar() energy=%d", g_game.player_state.energy);
	--g_game.player_state.colliding;
	if (g_game.player_state.energy) {
		--g_game.player_state.energy;
		return;
	}
	if (g_game.player_state.dead_flag) {
		return;
	}
	--g_game.player_state.lifes_count;
	if (g_game.player_state.lifes_count == 0) {
		g_game.quit_level_flag = 1;
	}
	g_game.player_state.tilemap_offset = g_game.player_state.next_tilemap_offset - _screen_cdg_tile_map_h - 2;
	g_game.fix_player_pos_tilemap_offset();	
	g_game.player_state.dead_flag = 1;
	g_game.update_nicky_anim_flag = 0;
	g_game.player_state.unk3D = 0;
	g_game.player_state.move_offsets_data_ptr = &move_offsets_table5[2];
	if (g_game.player_state.hdir_flags & 1) {
		g_game.player_state.anim_frame_ptr = &carlos_anim_frames_table__v3[24];
	} else {
		g_game.player_state.anim_frame_ptr = &carlos_anim_frames_table__v3[23];
	}
	g_game.player_state.anim_frames_count = 2;
	snd_play_sfx(&SAMPLES_SND[8 * 16]);
}

static void game_update_power_bar() {
#if 0 // v3
#endif
}

void game_run_cycle() {
	++g_game.cycle_counter;
	g_game.cycle_counter &= 1;
	if (g_game.cycle_counter == 0) {
		++g_game.cycle_counter2;
		g_game.cycle_counter2 &= 1;
	}
//	g_game._render->clearSpriteList();
//	inp_copy_state(g_game._render->get_keys_pressed_mask());
	game_adjust_objects_position();
	game_update_boss_object();
	g_game.update_carlos_object();
	g_game.update_oscar_object();
	game_handle_objects7();
	game_play_sound();
	game_adjust_tilemap();
	game_update_objects_bbox();
	game_execute_logic_op();
	game_draw_bonuses();
	game_update_objects2();
	game_execute_colliding_op();
	game_execute_colliding2_op();
	game_draw_carlos();
	game_update_tile_offsets();
	game_update_score();
	game_update_life_bar();
	game_update_power_bar();
	game_redraw_inventory();
	game_handle_level_change();
}

void Game::init_player_state() {
	memset(&player_state, 0, sizeof(player_state));
	player_state.dim_w = 16;
	player_state.dim_h = 61;
	player_state.pos_dx_1 = 14;
	player_state.pos_dy_1 = 60;
	player_state.anim_frame_ptr = &carlos_anim_frames_table__v3[19];
	player_state.hdir_flags = 1;
	player_state.vdir_flags = 0;
	player_state.pos_x = -16;
	player_state.pos_y = -16;
	player_state.lifes_count = 4;
	memset(&_oscarState, 0, sizeof(_oscarState));
	_oscarState.unk0 = 255;
	const int x = player_state.pos_x + 16;
	const int y = player_state.pos_y + 28;
	for (int i = 0; i < 20; ++i) {
		_oscarPosTable[i * 2] = x;
		_oscarPosTable[i * 2 + 1] = y;
	}
	g_game.carlos_move_offsets_ptr = (const uint16 *)&move_data[4];
}

void game_play_sound() {
	if (_snd_current_sfx_num >= 0) {
		snd_play_sfx(&SAMPLES_SND[_snd_current_sfx_num * 16]);
	}
	_snd_current_sfx_num = -1;
	_snd_current_sfx_priority = 0;
}

void game_draw_object(object_state_t *os) {
	if (g_game.monster_fight_flag > 2) {
		return;
	}
	const anim_data_t *ad = os->anim_data1_ptr;
	if (ad->anim_h == 0) {
		return;
	}
	const int x = os->pos_x + ad->draw_dx;
	const int y = os->pos_y + ad->draw_dy + 2;
//	const bool blinkSprite = (os->transparent_flag & 0x80) != 0 || (os->transparent_flag != 0 && g_game.cycle_counter == 0);
	if (ad->sprite_flags != 0x80) {
		g_game._render->addToSpriteList(g_game._res._boss_spr, ad->sprite_num, x, y);
	} else {
		g_game._render->addToSpriteList(g_game._res._level_spr, ad->sprite_num, x, y);
	}
	os->transparent_flag = 0;
}

void game_draw_carlos() {
	if (g_game.player_state.shield_duration != 0 && g_game.cycle_counter == 0) {
		// blinking
		return;
	}
	const anim_frame_t *af = g_game.player_state.anim_frame_ptr;
	const int y = g_game.player_state.pos_y + 2 + af->draw_dy;
	const int x = g_game.player_state.pos_x + af->draw_dx;
	const int num = af->sprite_num;
	if (num < 48) {
		g_game._render->addToSpriteList(g_game._res._carlos1_spr, num, x, y);
	} else {
		g_game._render->addToSpriteList(g_game._res._carlos2_spr, num - 48, x, y);
	}
}

static void game_draw_inventory_item(int item, int count) {
	const int y = 13;
	const int x = 5 + item * 18;
	static const int icon_map_table[] = { 14, 15, 17, 20 };
	g_game._render->addToSpriteList(g_game._res._level_spr, icon_map_table[item], x, y);
	g_game._render->addToSpriteList(DIGITS_SPR, count, x + 10, y + 8, 15);
}

static void game_draw_number(int x, int y, int n) {
	do {
		g_game._render->addToSpriteList(DIGITS_SPR, n % 10, x, y, 15);
		x -= 8;
		n /= 10;
	} while (n != 0);
}

static int _frameCounter, _framesPerSec;

void game_redraw_inventory() {
	game_draw_inventory_item(GAME_INV_LIFES, g_game.player_state.lifes_count);
	game_draw_inventory_item(GAME_INV_KEYS, g_game.player_state.keys_count);
	game_draw_inventory_item(GAME_INV_RAMS, g_game.player_state.rams_count);
	game_draw_inventory_item(GAME_INV_WINDS, g_game.player_state.winds_count);
	g_game._render->addToSpriteList(LIFEBAR_SPR, g_game.player_state.energy * 36 / 100, 4, 4, 1);
	game_draw_number(GAME_SCREEN_W - 16, 12, g_game.game_state.score);
	game_draw_number(GAME_SCREEN_W - 16, 20, g_game.player_state.energy);
	game_draw_number(GAME_SCREEN_W - 48, 28, g_game.player_state.pos_x);
	game_draw_number(GAME_SCREEN_W - 16, 28, g_game.player_state.pos_y);
	game_draw_number(GAME_SCREEN_W - 48, 36, g_game.decor_state.tile_map_offs_x / _screen_cdg_tile_map_h);
	game_draw_number(GAME_SCREEN_W - 16, 36, g_game.decor_state.tile_map_offs_y);
	++_frameCounter;
	if ((_frameCounter & 31) == 0) {
		static struct timeval t0, t1;
		gettimeofday(&t1, 0);
		const int msecs = (t1.tv_sec - t0.tv_sec) * 1000 + (t1.tv_usec - t0.tv_usec) / 1000;
		t0 = t1;
		if (msecs != 0) {
			_framesPerSec = (int)(1000. * 32 / msecs);
		}
	}
	game_draw_number(GAME_SCREEN_W - 16, 44, _framesPerSec);
}

static const uint8 unk16[] = {
	0x26, 0x46, 0x27, 0x48, 0x00, 0x00, 0x00, 0x00, 0x18, 0x2E,
	0x00, 0x00, 0x82, 0x42, 0x00, 0x00, 0x18, 0x2E, 0x00, 0x00,
	0x82, 0x42, 0x00, 0x00, 0x4E, 0x38, 0x4F, 0x3A, 0xDA, 0xDC,
	0xDB, 0xE1, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

uint8 game_set_decor_tile_flags4_3_2_0(uint8 b, int offset) {
	int ref_offset = g_game._res._decor_cdg[offset] * 8;
	if (g_game._res._decor_ref[ref_offset + 4] != 0) {
		++g_game.player_state.decor_ref_flag4;
	}
	if (g_game._res._decor_ref[ref_offset + 3] != 0) {
		++g_game.player_state.decor_ref_flag3;
	}
	if (g_game._res._decor_ref[ref_offset + 2] != 0 && g_game._res._decor_ref[ref_offset + 7] == 0) {
		g_game.player_state.decor_ref_flag2 = 1;
		const uint8 d = g_game._res._decor_cdg[offset];
		g_game._res._decor_cdg[offset] = g_game._res._decor_ref[ref_offset + 2];
		g_game.test_object_bounding_box_offset = -1;
		int num = g_game._currentLevel * 8;
		for (int i = 0; i < 2; ++i) {
			if (unk16[num] == 0) {
				break;
			}
			++num;
			if (unk16[num - 1] == d) {
				uint8 mask = unk16[num];
				if (mask != 0) {
					g_game._res._decor_cdg[offset] = mask;
				}
				mask = unk16[num + 2];
				if (mask != 0) {
					g_game._res._decor_cdg[offset + _screen_cdg_tile_map_h] = mask;
				}
				break;
			}
			num += 2;
			if (unk16[num - 1] == d) {
				uint8 mask = unk16[num];
				if (mask != 0) {
					g_game._res._decor_cdg[offset] = mask;
				}
				mask = unk16[num -  2];
				if (mask != 0) {
					g_game._res._decor_cdg[offset - _screen_cdg_tile_map_h] = mask;
				}
				break;
			}
			++num;
		}
	}
	const uint8 m = g_game._res._decor_ref[ref_offset];
	if (b & m) {
		g_game.player_state.decor_ref_flag0 |= b;
	}
	return m;
}

void game_set_decor_tile_flag0(uint8 b, int offset) {
	const int refOffset = g_game._res._decor_cdg[offset] * 8;
	if (b & g_game._res._decor_ref[refOffset]) {
		g_game.player_state.decor_ref_flag0 |= b;
	}
}

uint8 game_get_decor_tile_flag3(int offset) {
	uint8 flag = 0;
	if (offset > 0) {
		const int refOffset = g_game._res._decor_cdg[offset - 1] * 8;
		flag = g_game._res._decor_ref[refOffset + 3];
	}
	return flag;
}

uint16 game_get_decor_tile_flag7(int offset) {
	int refOffset = g_game._res._decor_cdg[offset] * 8;
	uint16 flag = g_game._res._decor_ref[refOffset + 7] & 7;
	if (flag != 0) {
		flag = (flag & 1) + 2;
		if ((g_game._res._decor_ref[refOffset + 7] & 4) != 0) {
			flag |= 0x8000;
		}
	}
	return flag;
}

void game_reset_objects8() {
	g_game.player_state.fire_button_counter = 0;
	g_game.bonus_count = 0;
	for (int i = 0; i < 17; ++i) {
		object_state_t *os = &g_game.objects_table_ptr8[i];
		os->displayed = 0;
		os->life = 0;
		os->distance_pos_x = 0;
		os->move_data_ptr = 0;
		os->anim_data_cycles = 0;
		os->unk3 = 0;
		os->distance_pos_y = 0;
		os->transparent_flag = 0;
	}
}

static void game_set_tile_offsets() {
	for (int i = 0; i < 256; ++i) {
		_screen_cdg_offsets[i] = i;
	}
	int tileNum = 1;
	for (int i = 0; i < 6; ++i) {
		tile_anim_t *ta = &g_game.game_state.tile_anims_table[i];
		ta->tile_num = tileNum;
		ta->anim = 0;
		for (int num = 0; num < 4; ++num) {
			ta->tiles_table[num] = _screen_cdg_offsets[tileNum];
			++tileNum;
		}
	}
	g_game.game_state.tile_cycling_counter = 0;
	int offset = g_game._currentLevel * 8;
	for (int i = 0; i < 4; ++i) {
		int num1 = unk16[offset++];
		int num2 = unk16[offset++];
		tile_anim_t *ta = &_screen_blinking_tile[i];
		ta->tile_num = num1;
		ta->anim = 0;
		ta->tiles_table[0] = _screen_cdg_offsets[num1];
		ta->tiles_table[1] = _screen_cdg_offsets[num2];
	}
}

static void game_update_cur_objects_ptr() {
	g_game.objects_list_cur = g_game.objects_list_head;
	g_game.objects_list_last = g_game.objects_list_head;
	const int x1 = _screen_tile_map_offs_x_shl4 - 128;
	const int x2 = _screen_tile_map_offs_x_shl4 + GAME_SCREEN_W + 96;
	for (object_state_t *os = g_game.objects_list_cur; (os->displayed & 0x80) == 0; ++os) {
		if (os->displayed != 0) {
			anim_data_t *ad = os->anim_data2_ptr;
			if (!ad) {
				warning("game_update_cur_objects_ptr() os->anim_data2_ptr is 0");
			} else if (ad->logic_opcode >= 21 && ad->logic_opcode <= 24) {
				os->tile_num = 0;
			}
			os->displayed = 1;
		}
	}
	for (object_state_t *os = g_game.objects_list_cur; (os->displayed & 0x80) == 0 && x1 > os->map_pos_x; ) {
		++g_game.objects_list_cur;
		os = g_game.objects_list_cur;
	}
	for (object_state_t *os = g_game.objects_list_cur; (os->displayed & 0x80) == 0 && x2 >= os->map_pos_x; ++os) {
		if (os->anim_data2_ptr == 0) {
			os->displayed = 0;
		}
		if (os->displayed != 0) {
			anim_data_t *ad = os->anim_data2_ptr;
			os->displayed = 2;
			os->pos_x = os->map_pos_x - _screen_tile_map_offs_x_shl4;
			os->pos_y = os->map_pos_y - _screen_tile_map_offs_y_shl4;
			os->anim_data1_ptr = ad;
			os->anim_data3_ptr = ad;
			os->anim_data4_ptr = ad->anim_data2_ptr;
			os->distance_pos_x = ad->distance_dx;
			os->distance_pos_y = ad->distance_dy;
			os->move_data_ptr = ad->move_data_ptr;
			os->anim_data_cycles = ad->cycles;
			os->unk3 = ad->unk3;
			os->life = ad->lifes;
			os->unk26 = ad->unk4;
			os->unk27 = ad->unk5;
			os->transparent_flag = 0;
			os->unk2C &= 0xFF00;
		}
		g_game.objects_list_last = os;
	}
}

static void game_init_level_helper2() {
	if (g_game.monster_fight_flag == 0) {
		g_game.update_nicky_anim_flag = 0;
		memset(&g_game.decor_state, 0, sizeof(g_game.decor_state));
		g_game.decor_state.tile_map_offs_x = _screen_tile_origin_x;
		_screen_tile_map_offs_x = _screen_tile_origin_x / _screen_cdg_tile_map_h;
		_screen_tile_map_offs_x_shl4 = _screen_tile_map_offs_x << 4;
		g_game.decor_state.tile_map_offs_y = _screen_tile_origin_y;
		_screen_tile_map_offs_y = _screen_tile_origin_y;
		_screen_tile_map_offs_y_shl4 = _screen_tile_map_offs_y << 4;
		g_game.decor_state.tile_map_end_offs_x = (_screen_cdg_tile_map_w - (GAME_SCREEN_W / 16) + 1) * _screen_cdg_tile_map_h;
		g_game.decor_state.tile_map_end_offs_y =  _screen_cdg_tile_map_h - (GAME_SCREEN_H / 16);
		g_game.decor_state.tile_block_x = g_game.decor_state.tile_map_offs_x;
		g_game.decor_state.tile_block_y = g_game.decor_state.tile_map_offs_y;
		game_set_tile_offsets();
		game_update_cur_objects_ptr();
		game_reset_objects8();
	}
}

void game_init_level_start_screen() {
	game_init_level_helper2();
	if (g_game.player_state.dead_flag) {
		game_init_carlos();
	}
	g_game.player_state.pos_x = g_game.player_pos_x - _screen_tile_map_offs_x_shl4;
	g_game.player_state.pos_y = g_game.player_pos_y - _screen_tile_map_offs_y_shl4;
	g_game.player_state.pos_dx_2 = 0;
	g_game.player_state.pos_dy_2 = 0;
	g_game.init_oscar_state();
	if ((g_game.player_state.hdir_flags & 1) == 0) {
		g_game.player_state.anim_frame_ptr = &carlos_anim_frames_table__v3[22];
	} else {
		g_game.player_state.anim_frame_ptr = &carlos_anim_frames_table__v3[19];
	}
	g_game.player_state.hdir_flags &= 1;
	game_redraw_inventory();
	if (g_game.monster_fight_flag == 0) {
#if 0 // v3
		for (int i = 0; i < 2; ++i) {
			g_game._render->clearSpriteList();
			game_adjust_tilemap();
			game_execute_logic_op();
			g_game._render->display_sprite_list_clipy_159();
		}
		g_game._render->clearSpriteList();
		screen_set_pal(_screen_pal_current);
#endif
	}
	++g_game.update_nicky_anim_flag;
}

void game_init_ref_ref(anim_data_t *ad, uint16 flags) {
	flags >>= 8;
	while (ad->unk0 >= 0) {
		ad->sprite_flags |= flags;
		++ad;
	}
	// pointer conversions of the original code is handled in load_ref_ref()
}

static void game_clear_object_table(object_state_t *os, int num) {
	for (int i = 0; i < num; ++i, ++os) {
		os->displayed = 0;
	}
}

void game_init_objects_from_positref(const uint8 *posit_ref, uint16 objects_offset) {
	game_clear_object_table(g_game.objects_table_ptr1, g_game._res._posit_count);
	game_clear_object_table(g_game.objects_table_ptr8, 10);
	game_clear_object_table(g_game.objects_table_ptr7, 2);
	game_clear_object_table(g_game.objects_table_ptr2, 7);
	g_game.objects_list_head = g_game.objects_table_ptr1;
	object_state_t *os = g_game.objects_list_head;
	while (1) {
		uint16 num = READ_LE_UINT16(posit_ref);
		if (num & 0x8000) {
			break;
		}
		anim_data_t *ad = &g_game._res._ref_ref[num];
		assert(os < g_game.objects_table_ptr2);
		os->anim_data1_ptr = ad;
		os->anim_data2_ptr = ad;
		os->anim_data3_ptr = ad;
		os->displayed = ad->unk0;
		os->life = ad->lifes;
		os->anim_data_cycles = ad->cycles;
		os->unk3 = ad->unk3;
		os->move_data_ptr = ad->move_data_ptr;
		os->distance_pos_x = ad->distance_dx;
		os->distance_pos_y = ad->distance_dy;
		os->anim_data4_ptr = ad->anim_data2_ptr;
		os->unk26 = ad->unk4;
		os->unk27 = ad->unk5;
		os->ref_ref_index = READ_LE_UINT16(posit_ref + 6);
		os->tile_num = posit_ref[9];
		os->visible = posit_ref[8];
		os->transparent_flag = 0;
		os->unk29 = 0;
		os->unk2C = 0;
		os->unk2E = 0;
		os->map_pos_x = 16 * (int16)READ_LE_UINT16(posit_ref + 2);
		os->map_pos_y = 16 * (1 + (int16)READ_LE_UINT16(posit_ref + 4));
		if (ad->sprite_num != 0 || ad->sprite_flags != 0) {
			int spr_w, spr_h;
			g_game._res.get_sprite_dim((ad->sprite_flags == 0x80) ? g_game._res._level_spr : g_game._res._boss_spr, ad->sprite_num, &spr_w, &spr_h);
			os->map_pos_y -= spr_h;
		}
		posit_ref += 10;
		++os;
	}
	os->displayed = 0xFF;
	os->map_pos_x = 30000;
}

void game_change_decor_tile_map_line(const object_state_t *os, const anim_data_t *ad) {
	const int x = os->map_pos_x + ad->anim_w / 2;
	const int y = os->map_pos_y + ad->anim_h - 8;
	const int offset = ((x >> 4) - 1) * _screen_cdg_tile_map_h + (y >> 4);
	for (int i = 0; i < os->tile_num; ++i) {
		g_game._res._decor_cdg[offset - i] = os->ref_ref_index;
	}
}

void game_change_decor_tile_map(const object_state_t *os) {
	const int offs = (os->map_pos_x >> 4) * _screen_cdg_tile_map_h + (os->map_pos_y >> 4);
	g_game._res._decor_cdg[offs] = os->tile_num;
}

void game_update_score() {
	if (g_game.cycle_score != 0) {
		g_game.game_state.extra_life_score += g_game.cycle_score;
		if (g_game.game_state.extra_life_score > 200000) {
			if (g_game.player_state.lifes_count < 9) {
				++g_game.player_state.lifes_count;
			}
			g_game.game_state.extra_life_score -= 200000;
			snd_play_sfx(&SAMPLES_SND[15 * 16]);
		}
		g_game.game_state.score += g_game.cycle_score;
		if (g_game.game_state.score > 9999999) {
			g_game.game_state.score = 9999999;
		}
	}
	g_game.cycle_score = 0;
}

void game_adjust_tilemap() {
	/* use 320 as the screen width for monster position */
	int monster_screen_x2 = _screen_cdg_tile_map_w - MAX(320, GAME_SCREEN_W) / 16;
	int monster_screen_y2 = _screen_cdg_tile_map_h - GAME_SCREEN_H / 16 - 1;
	_screen_tile_map_offs_y_shl4 = (g_game.decor_state.tile_map_offs_y << 4) - g_game.decor_state.scroll_dy;
	_screen_tile_map_offs_x_shl4 = ((g_game.decor_state.tile_map_offs_x / _screen_cdg_tile_map_h) << 4) - g_game.decor_state.scroll_dx;
	g_game._render->setTileMapOrigin(_screen_tile_map_offs_x_shl4, _screen_tile_map_offs_y_shl4);
	if (g_game.monster_fight_flag < 2) {
		if (_screen_tile_map_offs_x_shl4 >= (monster_screen_x2 << 4)) {
			if (_screen_tile_map_offs_y_shl4 >= (monster_screen_y2 << 4)) {
				++g_game.monster_fight_flag;
			}
		}
	}
	if (g_game.monster_fight_flag == 1) {
		game_setup_monster(0);
	} else if (g_game.monster_fight_flag == 3) {
		++g_game.monster_fight_flag;
		g_game.update_nicky_anim_flag = 0;
		const uint8 *p = g_game._res._posit_ref;
		for (int i = 0; i < 2; ++i) {
			while (READ_LE_UINT16(p) != 0xFFFF) {
				p += 10;
			}
			p += 10;
		}
		g_game.bossState.pos_y = ((READ_LE_UINT16(p + 4) + 1) << 4) - _screen_tile_map_offs_y_shl4;
		g_game.bossState.pos_x = (READ_LE_UINT16(p + 2) << 4) - _screen_tile_map_offs_x_shl4;
		g_game.bossState.frame_counter = 0;
		g_game.bossState.frame_num = 0;
		g_game.bossState.last_frame_num = READ_LE_UINT16(p + 6);
		g_game.bossState.frame_counter_mask = READ_LE_UINT16(p + 8);
//		_res_boss_spr = _res_friend_spr; // TODO: in draw_sprite, check boss_flag == 3
		g_game.game_state.boss_fight_counter = 0;
		g_game.game_state.boss_explosion_counter = 0;
		g_game.update_nicky_anim_flag = 255;
		game_run_cycle();
	}
}

void game_handle_level_change() {
	if (g_game.restart_level_flag) {
		debug(DBG_GAME, "game_handle_level_change() restart_level_flag=%d", g_game.restart_level_flag);
		g_game.restart_level_flag = 0;
		game_init_level_start_screen();
	} else {
		if (g_game.player_state.tile_blinking_duration != 0) {
			--g_game.player_state.tile_blinking_duration;
			for (int i = 0; i < 4; ++i) {
				tile_anim_t *ta = &_screen_blinking_tile[i];
				++ta->anim;
				if (ta->anim > 1) {
					ta->anim = 0;
				}
				const int num = (g_game.player_state.tile_blinking_duration == 0) ? 0 : ta->anim;
				_screen_cdg_offsets[ta->tile_num] = ta->tiles_table[num];
			}
		}
		if (g_game.player_state.has_spring > 0) {
			--g_game.player_state.has_spring;
			if (g_game.player_state.has_spring == 0) {
				g_game.carlos_move_offsets_ptr = (const uint16 *)&move_data[4];
			}
		}
		if (g_game.game_state.boss_fight_counter >= 60) {
			++g_game.game_state.boss_fight_counter;
			if (g_game.game_state.boss_fight_counter >= 200) {
				g_game.new_level_flag = 1;
			}
		}
	}
}

void game_execute_colliding_op() {
	if (g_game.player_state.dead_flag) {
		return;
	}
	int offset = ((g_game.player_state.pos_x + 12) >> 4) * _screen_cdg_tile_map_h + g_game.decor_state.tile_map_offs_x;
	if (g_game.player_state.pos_y >= 0 && g_game.player_state.pos_y < 160) {
		offset += ((g_game.player_state.pos_y + g_game.player_state.dim_h) >> 4) + g_game.decor_state.tile_map_offs_y;
		const int d[] = { 0, -1, -1, -1, _screen_cdg_tile_map_h, 1, 1, 1 };
		for (int i = 0; i < 8; ++i) {
			offset += d[i];
			game_collide_bonus(offset);
		}
	}
	g_game.player_state.colliding_flag = 0;
	if (g_game.boss_colliding_op_flag == 0) {
		if ((g_game.player_state.hdir_flags & 0x40) != 0) {
			g_game.bounding_box_y1 = g_game.player_state.pos_y + 32;
			g_game.bounding_box_x1 = g_game.player_state.pos_x + 11;
			g_game.bounding_box_y2 = g_game.bounding_box_y1 + 29;
			g_game.bounding_box_x2 = g_game.bounding_box_x1 + 26;
		} else {
			g_game.bounding_box_y1 = g_game.player_state.pos_y + 4;
			g_game.bounding_box_x1 = g_game.player_state.pos_x + 12;
			g_game.bounding_box_y2 = g_game.bounding_box_y1 + 50;
			g_game.bounding_box_x2 = g_game.bounding_box_x1 + 24;
		}
	} else if (g_game.boss_colliding_op_flag == 2) {
		g_game.bounding_box_y1 = g_game.player_state.pos_y + 28;
		g_game.bounding_box_x1 = g_game.player_state.pos_x + 5;
		g_game.bounding_box_y2 = g_game.bounding_box_y1 + 33;
		g_game.bounding_box_x2 = g_game.bounding_box_x1 + 38;
	} else {
		return;
	}
	game_execute_colliding_op_helper1();
	game_execute_colliding_op_helper2();
	for (object_state_t *os = g_game.objects_list_cur; os < g_game.objects_list_last; ++os) {
		if (os->displayed == 2 && (os->unk26 & 2) != 0) {
			anim_data_t *ad = os->anim_data1_ptr;
			if (game_test_object_bounding_box(os, ad, g_game.bounding_box_x1, g_game.bounding_box_y1, g_game.bounding_box_x2, g_game.bounding_box_y2)) {
				game_collides_op(os, ad);
			}
		}
	}
	if (g_game.player_state.colliding_flag != 0) {
		g_game.player_state.unk1 = 2;
		g_game.player_state.hdir_flags &= ~4;
		g_game.player_state.move_offsets_data_ptr = &g_game.carlos_move_offsets_ptr[3];
	}
}

void game_execute_colliding_op_helper1() {
	g_game.update_decor_counter = 7;
	game_handle_projectiles_collisions_helper2(g_game.objects_table_ptr2, 0, 7);
	if (g_game.update_decor_counter == 0) {
		return;
	}
	for (int i = 0; i < 7; ++i) {
		object_state_t *os = &g_game.objects_table_ptr2[i];
		if (os->displayed) {
			anim_data_t *ad = os->anim_data1_ptr;
			if (ad->unk4 & 1) {
				if (!game_test_object_bounding_box(os, ad, g_game.bounding_box_x1, g_game.bounding_box_y1, g_game.bounding_box_x2, g_game.bounding_box_y2)) {
					continue;
				}
				if (g_game.boss_colliding_op_flag == 0) {
					g_game.player_state.colliding += 4;
				}
				if (g_game.monster_fight_flag != 0) {
					g_game.player_state.colliding += 4;
				}
				if (ad->colliding_opcode == 42) {
					g_game.player_state.colliding += 8;
				}
				if (ad->anim_data3_ptr != 0) {
					game_init_object_from_anim3(os, ad);
				} else {
					os->displayed = 0;
				}
				break;
			}
		}
	}
}

void game_execute_colliding_op_helper2() {
	if (g_game.bonus_count == 0) {
		return;
	}
	for (int i = 0; i < 10; ++i) {
		object_state_t *os = &g_game.objects_table_ptr8[i];
		if (os->displayed != 0 && (os->unk26 & 2) != 0) {
			anim_data_t *ad = os->anim_data1_ptr;
			if (!game_test_object_bounding_box(os, ad, g_game.bounding_box_x1, g_game.bounding_box_y1, g_game.bounding_box_x2, g_game.bounding_box_y2)) {
				continue;
			}
			if (ad->colliding_opcode != 14) {
				os->transparent_flag |= 1;
			} else {
				g_game.cycle_score += ad->score;
				if (ad->anim_data3_ptr != 0) {
					game_play_anim_data_sfx(game_init_object_from_anim3(os, ad));
				} else {
					os->displayed = 0;
				}
			}
			break;
		}
	}
}

void game_execute_colliding2_op() {
	if (g_game.player_state.dead_flag) {
		return;
	}
	int offset = (g_game._oscarState.pos_x >> 4) * _screen_cdg_tile_map_h + g_game.decor_state.tile_map_offs_x;
	if (g_game._oscarState.pos_y + 6 >= 0 && g_game._oscarState.pos_y + 6 < 160) {
		offset += (g_game._oscarState.pos_y >> 4) + g_game.decor_state.tile_map_offs_y;
		game_collide_bonus(offset);
		offset += _screen_cdg_tile_map_h;
		game_collide_bonus(offset);
	}
	g_game.bounding_box_y1 = g_game._oscarState.pos_y + 2;
	if (g_game.bounding_box_y1 < 0) {
		return;
	}
	g_game.bounding_box_x1 = g_game._oscarState.pos_x + 2;
	if (g_game.bounding_box_x1 < 0) {
		return;
	}
	g_game.bounding_box_y2 = g_game.bounding_box_y1 + 14;
	if (g_game.bounding_box_y2 > 160) {
		return;
	}
	g_game.bounding_box_x2 = g_game.bounding_box_x1 + 28;
	if (g_game.bounding_box_x2 > GAME_SCREEN_W) {
		return;
	}
	game_execute_colliding_op_helper2();
	for (object_state_t *os = g_game.objects_list_cur; os < g_game.objects_list_last; ++os) {
		if (os->displayed == 2 && (os->unk26 & 2) != 0) {
			anim_data_t *ad = os->anim_data1_ptr;
			if (game_test_object_bounding_box(os, ad, g_game.bounding_box_x1, g_game.bounding_box_y1, g_game.bounding_box_x2, g_game.bounding_box_y2)) {
				game_collides_op_oscar(os, ad);
			}
		}
	}
}

void game_play_anim_data_sfx(const anim_data_t *ad) {
	if (ad->sound_num >= 0) {
		const uint8 priority = SAMPLES_SND[ad->sound_num * 16 + 12];
		if (priority >= _snd_current_sfx_priority) {
			_snd_current_sfx_priority = priority;
			_snd_current_sfx_num = ad->sound_num;
		}
	}
}

void game_setup_monster(int flag) {
	anim_data_t *ad = &g_game._res._ref_ref[0];
	while (ad->unk0 != -1) {
		++ad;
	}
	++ad;
	anim_data_t *boss_ref_ref = ad;

	const uint8 *boss_posit_ref = g_game._res._posit_ref;
	while (READ_LE_UINT16(boss_posit_ref) != 0xFFFF) {
		boss_posit_ref += 10;
	}
	boss_posit_ref += 10;
	game_init_ref_ref(boss_ref_ref, 0x9000);
	game_init_objects_from_positref(boss_posit_ref, 0);
	game_update_cur_objects_ptr();
	g_game.game_state.boss_fight_counter = 0;
	g_game.game_state.boss_explosion_counter = 0;
	g_game.update_nicky_anim_flag = 1;
}

anim_data_t *game_init_object_from_anim3(object_state_t *os, anim_data_t *ad) {
	os->life = ad->lifes;
	return os->init_object_from_anim3(ad);
}

void game_handle_projectiles_collisions_helper2(object_state_t *os, int _bp, int count) {
	while (count--) {
		anim_data_t *ad = os->anim_data1_ptr;
		if (os->displayed && (ad->unk4 & 1)) {
			int y = os->pos_y + g_game.decor_state.offs_y + (ad->anim_h >> 1);
			int x = os->pos_x + g_game.decor_state.offs_x + (ad->anim_w >> 1);
			int offs = (x >> 4) * _screen_cdg_tile_map_h + g_game.decor_state.tile_block_x + (y >> 4) + g_game.decor_state.tile_block_y;
			int tile_offs = g_game._res._decor_cdg[offs] * 8;
			if (g_game._res._decor_ref[tile_offs] & 0x10) {
				if (g_game._res._decor_ref[tile_offs + 5] != 0 && _bp != 0) {
					g_game._res._decor_cdg[offs] = g_game._res._decor_ref[tile_offs + 5];
					const uint8 priority = SAMPLES_SND[0 * 16 + 12];
					if (priority >= _snd_current_sfx_priority) {
						_snd_current_sfx_priority = priority;
						_snd_current_sfx_num = 0;
					}
				}
				if (ad->anim_data3_ptr != 0) {
					game_init_object_from_anim3(os, ad);
				} else {
					--g_game.update_decor_counter;
					os->displayed = 0;
				}
				return;
			}
		} else {
			--g_game.update_decor_counter;
		}
		++os;
	}
}

anim_data_t *game_logic_op_helper1(object_state_t *os, anim_data_t *ad) { // RENAME: game_reset_object_to_anim_pos
	if (ad->cycles != 0 && --os->anim_data_cycles == 0) {
		anim_data_t *next = ad->anim_data1_ptr;
		if (next == &g_game._res._ref_ref_use_ptr3) {
			next = os->anim_data3_ptr;
		}
		if (next == 0) {
			os->displayed = 0;
			return ad;
		}
		ad = next;
		os->anim_data1_ptr = ad;
		os->distance_pos_x = ad->distance_dx;
		os->distance_pos_y = ad->distance_dy;
		os->anim_data_cycles = ad->cycles;
	}
	os->set_to_initial_pos(ad);
	return ad;
}

void game_logic_op_helper2(object_state_t *_di, anim_data_t *_si, anim_data_t *ad, uint16 move) { // RENAME: game_add_object_relative_pos_anim
	anim_data_t *_game_logic_op_helper2_temp_var = ad;
	object_state_t *bx = g_game.objects_table_ptr8;
	for (int i = 0; i < 10; ++i) {
		if (bx->displayed == 0) {
			bx->anim_data1_ptr = _si;
			bx->anim_data3_ptr = _si;
			bx->displayed = _si->unk0;
			bx->life = _si->lifes;
			bx->anim_data_cycles = _si->cycles;
			bx->unk3 = _si->unk3;
			bx->displayed = 2;
			bx->anim_data4_ptr = _si->anim_data2_ptr;
			bx->move_data_ptr = _si->move_data_ptr;
			bx->unk26 = _si->unk4;
			bx->unk27 = _si->unk5;
			bx->pos_x = _di->pos_x + _game_logic_op_helper2_temp_var->dx2;
			bx->pos_y = _di->pos_y + _game_logic_op_helper2_temp_var->dy2;
			bx->move_data_ptr += getRandomNumber(move) * 2;
			bx->distance_pos_y = 0;
			bx->distance_pos_x = getRandomNumber(5) - 2;
			g_game.bonus_count = 1;
			++_di->ref_ref_index;
			game_play_anim_data_sfx(_si);
			return;
		}
		++bx;
	}
	++_di->ref_ref_index;
}

static void game_execute_logic_op_helper1(int x1, int x2) {
	if (g_game.execute_op_logic_helper1_flag <= 0) {
		while (x1 <= g_game.objects_list_cur->map_pos_x && g_game.objects_list_head != g_game.objects_list_cur) {
			--g_game.objects_list_cur;
		}
	} else {
		while (g_game.objects_list_cur->displayed != 2 && x2 >= g_game.objects_list_cur->map_pos_x && x1 > g_game.objects_list_cur->map_pos_x && (g_game.objects_list_cur->displayed & 0x80) == 0) {
			++g_game.objects_list_cur;
		}
	}
	g_game.execute_op_logic_helper1_flag = 0;
}

static void game_execute_logic_op_helper2(object_state_t *os_cur, anim_data_t *ad) {
	anim_data_t *ad4 = ad->anim_data4_ptr;
	if (ad4) {
		for (int i = 0; i < 7; ++i) {
			object_state_t *os = &g_game.objects_table_ptr2[i];
			if (os->displayed == 0) {
				os_cur->unk3 = ad->unk3;
				os->anim_data1_ptr = ad4;
				os->pos_x = os_cur->pos_x + ad->dx2;
				os->pos_y = os_cur->pos_y + ad->dy2;
				os->displayed = ad4->unk0;
				os->life = ad4->lifes;
				os->distance_pos_x = ad4->distance_dx;
				os->distance_pos_y = ad4->distance_dy;
				os->move_data_ptr = ad4->move_data_ptr;
				os->anim_data_cycles = ad4->cycles;
				os->unk3 = ad4->unk3;
				os->unk26 = ad4->unk4;
				os->unk27 = ad4->unk5;
				game_play_anim_data_sfx(ad4);
				if ((ad4->unk4 & 0x10) == 0) {
					game_move_object(os);
				}
				break;
			}
		}
	}
}

static void game_execute_logic_op_helper3(object_state_t *os, anim_data_t *ad) {
	if (os->displayed != 3) {
		if (os->unk2C != 0) {
			if (os->unk2C < 0) {
				++os->unk2C;
			} else {
				--os->unk2C;
			}
		}
		if ((os->unk26 & 0x40) != 0) {
			g_game.boss_logic_op_flag = -1;
		}
		ad = game_logic_op(os, os->anim_data1_ptr);
		if (os->displayed == 0) {
			os->displayed = (os->visible != 0) ? 1 : 0;
			return;
		}
		if (ad->unk3 != 0 && --os->unk3 <= 0) {
			os->unk3 = 1;
			if ((os->unk26 & 0x20) || (os->pos_x <= 320 && os->pos_x >= -32 && os->pos_y <= 160 && os->pos_y >= 0)) {
				game_execute_logic_op_helper2(os, ad);
			}
		}
	}
	if (os->pos_x >= -128 && os->pos_x <= 416 && os->pos_y >= -128 && os->pos_y <= 224) {
		os->displayed = 2;
		game_draw_object(os);
	} else {
		os->displayed = 3;
	}
}

void game_execute_logic_op() {
	g_game.boss_logic_op_flag = 0;
	int decor_x1 = _screen_tile_map_offs_x_shl4 - 128;
	int decor_x2 = _screen_tile_map_offs_x_shl4 + 416;
	int decor_y1 = _screen_tile_map_offs_y_shl4 - 128;
	int decor_y2 = _screen_tile_map_offs_y_shl4 + 224;
	game_execute_logic_op_helper1(decor_x1, decor_x2);
	object_state_t *os = g_game.objects_list_cur;
	for (; (os->displayed & 0x80) == 0; ++os) {
		if (os->displayed == 0) {
			continue;
		}
		if (os->displayed >= 2) {
			game_execute_logic_op_helper3(os, 0);
			continue;
		}
		if (os >= g_game.objects_list_last) {
			if (decor_x2 < os->map_pos_x) {
				break;
			}
		} else {
			if (decor_x2 < os->map_pos_x) {
				continue;
			}
		}
		if (decor_x1 > os->map_pos_x || decor_y2 < os->map_pos_y || decor_y1 > os->map_pos_y || os->displayed == 0) {
			continue;
		}
		if (os->anim_data2_ptr == 0) {
			os->displayed = 0;
		} else {
			anim_data_t *ad = os->anim_data2_ptr;
			if (os > g_game.objects_list_last) {
				g_game.objects_list_last = os;
			}
			os->pos_y = os->map_pos_y - _screen_tile_map_offs_y_shl4;
			os->pos_x = os->map_pos_x - _screen_tile_map_offs_x_shl4;
			os->anim_data1_ptr = ad;
			os->anim_data3_ptr = ad;
			os->anim_data4_ptr = ad->anim_data2_ptr;
			os->distance_pos_x = ad->distance_dx;
			os->move_data_ptr = ad->move_data_ptr;
			os->displayed = ad->unk0;
			os->life = ad->lifes;
			os->distance_pos_y = ad->distance_dy;
			os->anim_data_cycles = ad->cycles;
			os->unk3 = ad->unk3;
			os->displayed = 2;
			os->unk26 = ad->unk4;
			os->unk27 = ad->unk5;
			os->transparent_flag = 0;
			os->unk2C = 0;
			if (ad->logic_opcode >= 21 && ad->logic_opcode <= 24) {
				os->tile_num = 0;
			}
			game_execute_logic_op_helper3(os, ad);
		}
	}
	if (os <= g_game.objects_list_last && os != g_game.objects_list_head) {
		--g_game.objects_list_last;
	}
}

int game_get_decor_offset(int x, int y) {
	int offset = ((x + g_game.decor_state.offs_x - g_game.decor_state.delta_x) >> 4) * _screen_cdg_tile_map_h + g_game.decor_state.tile_block_x;
	offset += ((y + g_game.decor_state.offs_y - g_game.decor_state.delta_y) >> 4) + g_game.decor_state.tile_block_y;
	return offset;
}

int game_get_nicky_offset(int x, int y) {
	int offset = (x >> 4) * _screen_cdg_tile_map_h + g_game.decor_state.tile_map_offs_x;
	offset += (y >> 4) + g_game.decor_state.tile_map_offs_y;
	return offset;
}

int game_test_object_bounding_box(object_state_t *os, anim_data_t *ad, int x1, int y1, int x2, int y2) {
	int box_x1 = 0;
	int box_x2 = 0;
	int box_y1 = 0;
	int box_y2 = 0;
	if (ad) {
		box_x1 = ad->bounding_box_x1;
		box_x2 = ad->bounding_box_x2;
		box_y1 = ad->bounding_box_y1;
		box_y2 = ad->bounding_box_y2;
	}
	if (os->pos_y + box_y2 < y1 || os->pos_y + box_y1 > y2) {
		return 0;
	}
	if (os->pos_x + box_x2 < x1 || os->pos_x + box_x1 > x2) {
		return 0;
	}
	return 1;
}

int game_test_object_on_screen(object_state_t *os, anim_data_t *ad) {
	if (os->pos_y >= GAME_SCREEN_H || os->pos_y + ad->bounding_box_y2 < 0) {
		return 0;
	}
	if (os->pos_x >= GAME_SCREEN_W || os->pos_x + ad->bounding_box_x2 < 0) {
		return 0;
	}
	return 1;
}

void game_collide_bonus(int offset) {
	const int refOffset = g_game._res._decor_cdg[offset] * 8;
	const int num = g_game._res._decor_ref[refOffset + 7];
	const int type = (num >> 4) & 15;
	if (type == 0) {
		return;
	}
	g_game._res._decor_cdg[offset] = g_game._res._decor_ref[refOffset + 2];
	g_game.test_object_bounding_box_result = 0xFF;
	if (type >= 1 && type <= 10) {
		g_game.cycle_score += type * 10;
		_snd_current_sfx_num = 2;
	} else if (type == 15) {
		++g_game.winds_count;
		if (g_game.winds_count == 10) {
			g_game.winds_count = 0;
			if (g_game.player_state.winds_count < 9) {
				++g_game.player_state.winds_count;
			}
		}
		g_game.cycle_score += 250;
		_snd_current_sfx_num = 5;	
	}
}

void Game::init_oscar_state() {
	_oscarState.resetPosition(player_state.pos_x + 16, player_state.pos_y + 28, _oscarPosTable);
}

void Game::draw_oscar_object() {
	static const uint8 oscar_sprite_num[] = { 0, 1, 2, 3, 2, 1 };
	int spr_num = oscar_sprite_num[_oscarState.frame_num];
	if (_oscarState.hdir_flags & 2) {
		spr_num += ((_oscarState.hdir_flags & 1) != 0) ? 9 : 4;
	} else {
		spr_num += ((_oscarState.hdir_flags & 1) == 0) ? 5 : 0;
	}
	g_game._render->addToSpriteList(g_game._res._oscar_spr, spr_num, _oscarState.pos_x, _oscarState.pos_y);
}

void Game::update_oscar_object() {
	int x, y;
	if (g_game.player_state.dead_flag || g_game.player_state.pos_y > 120) {
		x = 144; // GAME_SCREEN_W / 2 - 16
		y = 70;
	} else {
		x = g_game.player_state.pos_x;
		y = g_game.player_state.pos_y - 16;
		if (g_game.player_state.hdir_flags & 0x40) {
			y += 32;
		}
		if (g_game.player_state.hdir_flags & 0x36) {
			y += 24;
		}
		if ((g_game.player_state.hdir_flags & 1) == 0) {
			x += 36;
		}
	}
	int current_x_pos = _oscarPosTable[38];
	int current_y_pos = _oscarPosTable[39];
	memmove(_oscarPosTable + 2, _oscarPosTable, sizeof(_oscarPosTable) - 2 * sizeof(int16));
	_oscarPosTable[0] = x;
	_oscarPosTable[1] = y;
	_oscarState.dx = 1;
	_oscarState.dy = 1;
	if (g_game.cycle_counter != 0) {
		++_oscarState.dx;
		++_oscarState.dy;
	}
	if (_oscarState.pos_y >= current_y_pos - 8) {
		if (_oscarState.pos_y > current_y_pos - 4) {
			_oscarState.dy = -_oscarState.dy;
		} else {
			_oscarState.dy = 0;
		}
	}
	if (_oscarState.pos_x >= current_x_pos - 8) {
		if (_oscarState.pos_x > current_x_pos - 4) {
			_oscarState.dx = -_oscarState.dx;
		} else {
			_oscarState.dx = 0;
		}
	}
	if ((_oscarState.hdir_flags & 2) == 0) {
		int x = _oscarPosTable[0];
		int y = _oscarPosTable[1];
		bool posChanged = false;
		for (int i = 0; i < 20; ++i) {
			if (x != _oscarPosTable[2 * i] || y != _oscarPosTable[2 * i + 1]) {
				posChanged = true;
				break;
			}
		}
		if (!posChanged && _oscarState.dx == 0 && _oscarState.dy == 0) {
			if ((g_game.player_state.hdir_flags & 8) == 0) {
				if ((g_game.player_state.hdir_flags & 1) != (_oscarState.hdir_flags & 1)) {
					_oscarState.resetDirection();
					draw_oscar_object();
					return;
				}
			}
		}
		if (_oscarState.dx < 0 && (_oscarState.hdir_flags & 1) != 0) {
			_oscarState.resetDirection();
			draw_oscar_object();
			return;
		}
		if (_oscarState.dx > 0 && (_oscarState.hdir_flags & 1) == 0) {
			_oscarState.resetDirection();
			draw_oscar_object();
			return;
		}
		_oscarState.pos_x += _oscarState.dx;
		_oscarState.pos_y += _oscarState.dy;
		int mask = 3;
		if (_oscarState.dx != 0 && _oscarState.dy != 0) {
			mask = 1;
		}
		++_oscarState.frame_counter;
		_oscarState.frame_counter &= mask;
		if (_oscarState.frame_counter == 0) {
			++_oscarState.frame_num;
			if (_oscarState.frame_num >= 6) {
				_oscarState.frame_num = 0;
			}
		}
	} else {
		_oscarState.dx = _oscarState.dy = 0;
		_oscarState.frame_counter = 0;
		_oscarState.frame_num = 0;
		_oscarState.hdir_flags = (~_oscarState.hdir_flags) & 1;
	}
	draw_oscar_object();
}

void Game::fix_player_pos_tilemap_offset() {
	const int offset = player_state.tilemap_offset - 2;
	int base_tile_pos_x = offset / _screen_cdg_tile_map_h;
	int base_tile_pos_y = offset % _screen_cdg_tile_map_h;
	if (base_tile_pos_y < 0) {
		base_tile_pos_y = 0;
	}
	int tile_origin_y = base_tile_pos_y - 2;
	if (tile_origin_y < 0) {
		tile_origin_y = 0;
	}
	const int max_tile_y = _screen_cdg_tile_map_h2 - 10;
	if (tile_origin_y > max_tile_y) {
		tile_origin_y = max_tile_y;
	}
	int tile_origin_x = base_tile_pos_x * _screen_cdg_tile_map_h - _screen_cdg_tile_map_h * 10;
	if (tile_origin_x < 0) {
		tile_origin_x = 0;
	}
	const int max_tile_x = (_screen_cdg_tile_map_w - 320 / 16) * _screen_cdg_tile_map_h;
	if (tile_origin_x < max_tile_x) {
		tile_origin_x += _screen_cdg_tile_map_h;
	} else {
		tile_origin_x = max_tile_x;
	}
	if (g_game.monster_fight_flag) {
		tile_origin_x = max_tile_x;
		tile_origin_y = max_tile_y;
	}
	_screen_tile_origin_x = tile_origin_x;
	_screen_tile_origin_y = tile_origin_y;
	player_pos_x = base_tile_pos_x * 16;
	player_pos_y = base_tile_pos_y * 16 + 1;
}

void game_update_objects_bbox() {
	for (object_state_t *os = g_game.objects_list_cur; os < g_game.objects_list_last; ++os) {
		if (os->displayed == 2 && os->unk26 == 1) {
			const anim_data_t *ad = os->anim_data1_ptr;
			g_game.bounding_box_y2 = os->pos_y + ad->bounding_box_y2;
			if (g_game.bounding_box_y2 < 0) {
				continue;
			}
			g_game.bounding_box_y1 = os->pos_y + ad->bounding_box_y1;
			if (g_game.bounding_box_y2 > 160) {
				continue;
			}
			g_game.bounding_box_x2 = os->pos_x + ad->bounding_box_x2;
			if (g_game.bounding_box_x2 < 0) {
				continue;
			}
			g_game.bounding_box_x1 = os->pos_x + ad->bounding_box_x1;
			if (g_game.bounding_box_x1 > 320) {
				continue;
			}
			g_game.ret_flag = 0;
			game_update_objects_bbox_helper1(os);
			if (g_game.ret_flag >= 0) {
				game_update_objects_bbox_helper2(os);
			}
		}
	}
}

void game_update_objects_bbox_helper1(object_state_t *os) {
	object_state_t *os_ptr = g_game.objects_table_ptr7;
	if (os_ptr->displayed != 0) {
		g_game.bounding_box_y1 = os_ptr->pos_y - 16;
		g_game.bounding_box_y2 = os_ptr->pos_y + 77;
		const int a = g_game.player_state.power * 2;
		g_game.test_object_bounding_box_offset = -1;
		if ((g_game.player_state.hdir_flags & 1) == 0) {
			g_game.test_object_bounding_box_offset = 0;
			g_game.bounding_box_x1 = g_game.bounding_box_x2 - a;
			g_game.bounding_box_x2 = os_ptr->pos_x + 24;
		} else {
			g_game.bounding_box_x1 = os_ptr->pos_x;
			g_game.bounding_box_x2 = g_game.bounding_box_x1 + a;
		}
		anim_data_t *ad = os->anim_data1_ptr;
		if (os->pos_y + ad->bounding_box_y2 < g_game.bounding_box_y1) {
			return;
		}
		if (os->pos_y + ad->bounding_box_y1 > g_game.bounding_box_y2) {
			return;
		}
		if (os->pos_x + ad->bounding_box_x2 < g_game.bounding_box_x1) {
			return;
		}
		if (os->pos_x + ad->bounding_box_x1 > g_game.bounding_box_x2) {
			return;
		}
		g_game.ret_flag = -1;
		if (ad->lifes == 0) {
			return;
		}
		os->transparent_flag = 0xFF;
		if (g_game.monster_fight_flag) {
			os->life -= g_game.cycle_counter;
		} else {
			--os->life;
		}
		if (os->life > 0) {
			switch (ad->logic_opcode) {
			case 1:
			case 5:
			case 6:
			case 12:
			case 13:
				if ((g_game.player_state.hdir_flags & 1) == 0) {
					os->unk2C = -7;
				} else {
					os->unk2C = 7;
				}
				break;
			}
			return;
		}
		if (ad->colliding_opcode == 10) {
			game_change_decor_tile_map_line(os, ad);
		}
		g_game.unk_ref_index = 0;
		g_game.cycle_score = ad->score;
		if ((ad->unk4 & 8) != 0) {
			g_game.update_object_tilemap(os);
		}
		if (!ad->anim_data3_ptr) {
			os->displayed = 0;
			if (!os->visible) {
				return;
			}
			os->displayed = 1;
			ad = os->anim_data1_ptr;
		} else {
			if (!os->visible) {
				os->anim_data2_ptr = 0;
			}
			os->transparent_flag = 0;
			ad = game_init_object_from_anim3(os, ad);
		}
		game_play_anim_data_sfx(ad);
	}
}

void game_update_objects_bbox_helper2(object_state_t *os) {
	object_state_t *_si = &g_game.objects_table_ptr7[1];
	if (_si->displayed == 0) {
		return;
	}
	g_game.bounding_box_y1 = _si->pos_y;
	g_game.bounding_box_y2 = _si->pos_y + 32;
	g_game.bounding_box_x1 = _si->pos_x;
	g_game.bounding_box_x2 = _si->pos_x + 32;
	anim_data_t *ad = os->anim_data1_ptr;
	if (os->pos_y + ad->bounding_box_y2 < g_game.bounding_box_y1) {
		return;
	}
	if (os->pos_y + ad->bounding_box_y1 > g_game.bounding_box_y2) {
		return;
	}
	if (os->pos_x + ad->bounding_box_x2 < g_game.bounding_box_x1) {
		return;
	}
	if (os->pos_x + ad->bounding_box_x1 > g_game.bounding_box_x2) {
		return;
	}
	g_game.ret_flag = -1;
	if (ad->lifes == 0) {
		return;
	}
	os->transparent_flag = 0xFF;
	os->life -= (g_game.monster_fight_flag == 0) ? 8 : 1;
	if (os->life > 0) {
		return;
	}
	if (ad->colliding_opcode == 10) {
		game_change_decor_tile_map_line(os, ad);
	}
	g_game.unk_ref_index = 0;
	g_game.cycle_score += ad->score;
	if (ad->unk4 & 8) {
		g_game.update_object_tilemap(os);
	}
	if (ad->anim_data3_ptr == 0) {
		os->displayed = 0;
		if (os->visible == 0) {
			return;
		}
		os->displayed = 1;
		game_play_anim_data_sfx(os->anim_data1_ptr);
	} else {
		if (os->visible == 0) {
			os->anim_data2_ptr = 0;
		}
		os->transparent_flag = 0;
		game_play_anim_data_sfx(game_init_object_from_anim3(os, ad));
	}
}

void game_update_boss_object() {
	if (g_game.monster_fight_flag == 0) {
		if (g_game.boss_logic_op_flag != 0 && g_game.boss_colliding_op_flag == 0) {
			g_game.colliding_boss_object_flag = 0;
			if (g_game.player_state.dead_flag != 0) {
				g_game.player_state.unk3D = 0;
				return;
			}
			if (g_game.player_state.unk1 != 2 || (g_game.player_state.move_offsets_data_ptr[0] & 0x8000) == 0) {
				const int x = g_game.player_state.pos_x + 18;
				const int y = g_game.player_state.pos_y + 61;
				g_game.bounding_box_y1 = y - 2;
				g_game.bounding_box_y2 = y + 2;
				g_game.bounding_box_x1 = x + 2;
				g_game.bounding_box_x2 = x + 14;
				for (object_state_t *os = g_game.objects_list_cur; os < g_game.objects_list_last; ++os) {
					if (os->displayed != 2 || (os->unk26 & 0x40) == 0) {
						continue;
					}
					const anim_data_t *ad = os->anim_data1_ptr;
					if (os->pos_y + ad->bounding_box_y2 < g_game.bounding_box_y1) {
						continue;
					}
					if (os->pos_y + ad->bounding_box_y1 > g_game.bounding_box_y2) {
						continue;
					}
					if (os->pos_x + ad->bounding_box_x2 < g_game.bounding_box_x1) {
						continue;
					}
					if (os->pos_x + ad->bounding_box_x1 > g_game.bounding_box_x2) {
						continue;
					}
					g_game.colliding_boss_object_flag = 1;
					g_game.player_state.pos_y += ad->bounding_box_y1 - g_game.player_state.dim_h + os->distance_pos_y;
					g_game.player_state.pos_x += os->distance_pos_x;
					g_game.player_state.hdir_flags &= ~0x34;
					g_game.player_state.unk1 = 0;
					g_game.player_state.move_offsets_data_ptr = (const uint16 *)&move_data[16];
					break;
				}
			}
		}
		g_game.colliding_boss_object = 0;
	} else if (g_game.monster_fight_flag >= 4) {
		if (g_game.monster_fight_flag == 4) {
			bool draw_sprite = true;
			++g_game.bossState.frame_counter;
			if (g_game.bossState.frame_counter >= 3) {
				g_game.bossState.frame_counter = 0;
				++g_game.bossState.frame_num;
				if (g_game.bossState.frame_num == g_game.bossState.last_frame_num) {
					g_game.monster_fight_flag = 5;
					draw_sprite = false;
				}
			}
			if (draw_sprite) {
				int boss_spr_w, boss_spr_h;
				g_game._res.get_sprite_dim(g_game._res._friend_spr, g_game.bossState.frame_num, &boss_spr_w, &boss_spr_h);
				const int x = g_game.bossState.pos_x;
				const int y = g_game.bossState.pos_y + 3 - boss_spr_h;
				g_game._render->addToSpriteList(g_game._res._friend_spr, g_game.bossState.frame_num, x, y);
				return;
			}
		}
		++g_game.bossState.frame_counter;
		g_game.bossState.frame_counter &= 15;
		const int frame_num = g_game.bossState.last_frame_num - 1 + ((g_game.bossState.frame_counter >> 3) & g_game.bossState.frame_counter_mask);
		int boss_spr_w, boss_spr_h;
		g_game._res.get_sprite_dim(g_game._res._friend_spr, frame_num, &boss_spr_w, &boss_spr_h);
		const int x = g_game.bossState.pos_x;
		const int y = g_game.bossState.pos_y + 3 - boss_spr_h;
		g_game._render->addToSpriteList(g_game._res._friend_spr, frame_num, x, y);
		if ((g_game.player_state.hdir_flags & 0x7F) == 0 && g_game.player_state.unk1 == 0 && (inp_direction_mask & 3) == 0) {
			g_game.bounding_box_y1 = g_game.player_state.pos_y + 4;
			g_game.bounding_box_x1 = g_game.player_state.pos_x + 15;
			g_game.bounding_box_y2 = g_game.bounding_box_y1 + 50;
			g_game.bounding_box_x2 = g_game.bounding_box_x1 + 18;
			if (g_game.bossState.pos_y + 3 < g_game.bounding_box_y1) {
				return;
			}
			if (g_game.bossState.pos_y + 3 - boss_spr_h > g_game.bounding_box_y2) {
				return;
			}
			if (g_game.bossState.pos_x + boss_spr_w < g_game.bounding_box_x1) {
				return;
			}
			if (g_game.bossState.pos_x > g_game.bounding_box_x2) {
				return;
			}
			g_game.new_level_flag = 1;
		}
	}			
}


static void game_handle_objects7_helper1(object_state_t *os) {
	if (g_game.player_state.unk1 != 0 || (g_game.player_state.hdir_flags & 0xFE) != 0) {
		g_game.player_state.has_blue_ball |= 1;
	} else {
		g_game.player_state.has_blue_ball &= ~1;
	}
	os->displayed = 1;
	if (g_game.player_state.dead_flag == 0) {
		snd_play_sfx(&SAMPLES_SND[24 * 16]);
	}
	++g_game.player_state.increment_power_bar;
	g_game.player_state.action_code = 1;
	g_game.player_state.hdir_flags &= 3;
	if ((g_game.player_state.hdir_flags & 1) != 0) {
		os->ref_ref_index = 10;
	} else {
		os->ref_ref_index = 15;
	}
}

static void game_handle_objects7_helper2() {
	object_state_t *os = g_game.objects_table_ptr7;
	if (os->displayed != 0) {
		if ((g_game.player_state.hdir_flags & 2) == 0) {
			if ((g_game.player_state.hdir_flags & 1) == 0) {
				g_game.player_state.anim_frame_ptr = &carlos_anim_frames_table__v3[22];
			} else {
				g_game.player_state.anim_frame_ptr = &carlos_anim_frames_table__v3[19];
			}
			g_game.player_state.anim_frames_count = 2;
			g_game.player_state.hdir_flags &= 1;
		}
		g_game.player_state.increment_power_bar = 255;
		os->displayed = 0;
	}
}

static void game_handle_objects7_helper3() {
	if (g_game.boss_colliding_op_flag == 1 && (g_game.player_state.anim_frame_ptr == &carlos_anim_frames_table__v3[54] || g_game.player_state.anim_frame_ptr == &carlos_anim_frames_table__v3[70])) {
		g_game.boss_colliding_op_flag = 2;
	}
	if (g_game.boss_colliding_op_flag < 0 || (g_game.player_state.anim_frame_ptr == &carlos_anim_frames_table__v3[19] || g_game.player_state.anim_frame_ptr == &carlos_anim_frames_table__v3[22])) {
		g_game.boss_colliding_op_flag = 0;
		g_game.player_state.hdir_flags &= 1;
	}
}

void game_handle_objects7() {
	object_state_t *os = g_game.objects_table_ptr7;
	if (g_game.boss_colliding_op_flag == 0) {
		if (inp_fire_button != 0) {
			if (os->displayed != 0) {
				if (g_game.player_state.power == 56) {
					if (g_game.player_state.winds_count == 0) {
						os->displayed = 0;
						g_game.player_state.increment_power_bar = 0xFF;
						if (g_game.player_state.has_blue_ball == 1) {
							if (g_game.player_state.hdir_flags & 1) {
								g_game.player_state.anim_frame_ptr = &carlos_anim_frames_table__v3[19];
							} else {
								g_game.player_state.anim_frame_ptr = &carlos_anim_frames_table__v3[22];
							}
							g_game.player_state.anim_frames_count = 2;
							g_game.player_state.hdir_flags &= 1;
						}
					} else {
						g_game.player_state.increment_power_bar = 255;
						os->displayed = 0;
						if ((g_game.player_state.hdir_flags & 1) == 0) {
							g_game.player_state.anim_frame_ptr = &carlos_anim_frames_table__v3[22];
						} else {
							g_game.player_state.anim_frame_ptr = &carlos_anim_frames_table__v3[19];
						}
						g_game.player_state.anim_frames_count = 2;
						g_game.player_state.hdir_flags &= 1;
						os = &g_game.objects_table_ptr7[1];
						if (os->displayed == 0 && g_game.player_state.winds_count != 0) {
							--g_game.player_state.winds_count;
							os->displayed = 1;
							os->map_pos_x = 0;
							os->map_pos_y = 1;
							os->distance_pos_x = 0;
							os->distance_pos_y = 0;
							os->ref_ref_index = 0;
							os->anim_data_cycles = 0;
							os->unk3 = 0;
						}
					}
				} else {
					++g_game.player_state.increment_power_bar;
				}
			} else if (g_game.player_state.power == 0 && (g_game.player_state.hdir_flags & 0x70) == 0) {
				if ((g_game.player_state.hdir_flags & 1) == 0) {
					int offset = (g_game.player_state.pos_x + g_game.player_state.pos_dx_2 + 8) >> 4;
					offset = offset * _screen_cdg_tile_map_h + g_game.decor_state.tile_block_x;
					g_game.player_state.decor_ref_flag0 = 0;
					if (g_game.player_state.pos_y <= 128) {
						offset += g_game.player_state.pos_y;
						offset += (g_game.player_state.pos_dy_2 + 10) >> 4;
						offset += g_game.decor_state.tile_block_y;
						game_set_decor_tile_flag0(8, offset);
						if ((g_game.player_state.decor_ref_flag0 & 8) == 0) {
							game_set_decor_tile_flag0(8, offset - _screen_cdg_tile_map_h);
							if ((g_game.player_state.decor_ref_flag0 & 8) == 0) {
								game_handle_objects7_helper1(os);
							}
						}
					}
				} else {
					int offset = (g_game.player_state.pos_x + g_game.player_state.pos_dx_2 + 40) >> 4;
					offset = offset * _screen_cdg_tile_map_h + g_game.decor_state.tile_block_x;
					g_game.player_state.decor_ref_flag0 = 0;
					if (g_game.player_state.pos_y <= 128) {
						offset += g_game.player_state.pos_y;
						offset += (g_game.player_state.pos_dy_2 + 10) >> 4;
						offset += g_game.decor_state.tile_block_y;
						game_set_decor_tile_flag0(4, offset);
						if ((g_game.player_state.decor_ref_flag0 & 4) == 0) {
							game_set_decor_tile_flag0(8, offset + _screen_cdg_tile_map_h);
							if ((g_game.player_state.decor_ref_flag0 & 4) == 0) {
								game_handle_objects7_helper1(os);
							}
						}
					}
				}
			}
		} else if (inp_fire_button == 0) {
			game_handle_objects7_helper2();
		}
		while (g_game.objects_table_ptr7[0].displayed != 0) {
			os = g_game.objects_table_ptr7;
			if ((g_game.player_state.hdir_flags & 8) != 0) {
				game_handle_objects7_helper2();
				continue;
			}
			int count = g_game.player_state.power >> 3;
			if (count > 0 && count < 6) {
				--count;
				int _temp_local1 = count;
				int _temp_local4 = count;
				int y = g_game.player_state.pos_y;
				int x = g_game.player_state.pos_x;
				if ((g_game.player_state.has_blue_ball & 1) == 0) {
					if ((g_game.player_state.hdir_flags & 1) == 0) {
						x -= 20;
					} else {
						x += 42;
					}
					y += 9;
					_temp_local1 += os->ref_ref_index;
				} else {
					if ((g_game.player_state.hdir_flags & 1) == 0) {
						x -= 10;
					} else {
						x += 30;
					}
					y += 7;
					_temp_local1 += os->ref_ref_index;
				}
				static const int8 word_1B00[] = { 3, 0, 0, 0, 0, 0, 0, 0, -4, 7, 0, 0, 0, -7, -7 };
				y += word_1B00[_temp_local4];
				if ((g_game.player_state.hdir_flags & 1) == 0) {
					x += word_1B00[5 + _temp_local4] + g_game.player_state.power;
				} else {
					x += word_1B00[10 + _temp_local4] - g_game.player_state.power;
				}
				y += 2;
				os->distance_pos_y = y;
				os->distance_pos_x = x;
				g_game._render->addToSpriteList(g_game._res._oscar_spr, _temp_local1, x, y);
				if ((g_game.player_state.hdir_flags & 1) == 0) {
					int offset = (os->distance_pos_x + 8) >> 4;
					offset = offset * _screen_cdg_tile_map_h + g_game.decor_state.tile_block_x;
					g_game.player_state.decor_ref_flag0 = 0;
					offset += ((os->distance_pos_x + 8) >> 4) + g_game.decor_state.tile_block_y;
					game_set_decor_tile_flag0(8, offset);
					if ((g_game.player_state.decor_ref_flag0 & 8) == 0) {
						game_set_decor_tile_flag0(8, offset - _screen_cdg_tile_map_h);
						if ((g_game.player_state.decor_ref_flag0 & 8) == 0) {
							break;
						}
					}
				} else {
					int offset = (os->distance_pos_x + 4) >> 4;
					offset = offset * _screen_cdg_tile_map_h + g_game.decor_state.tile_block_x;
					g_game.player_state.decor_ref_flag0 = 0;
					offset += ((os->distance_pos_x + 8) >> 4) + g_game.decor_state.tile_block_y;
					game_set_decor_tile_flag0(4, offset);
					if ((g_game.player_state.decor_ref_flag0 & 4) == 0) {
						game_set_decor_tile_flag0(4, offset + _screen_cdg_tile_map_h);
						if ((g_game.player_state.decor_ref_flag0 & 4) == 0) {
							break;
						}
					}
				}
				game_handle_objects7_helper2();
				continue;
			}
			break;
		}
	}
	os = &g_game.objects_table_ptr7[1];
	if (os->displayed != 0) {
		if (g_game.cycle_counter == 0) {
			g_game.player_state.colliding >>= 1;
		}
		++os->anim_data_cycles;
		if (os->anim_data_cycles > 2) {
			os->anim_data_cycles = 0;
		}
		const int angleInc = (os->ref_ref_index >> 1) & ~1;
		const int dy = g_game.player_state.pos_y + 14;
		const int dx = g_game.player_state.pos_x + 8;
		bool drawSprite = true;
		if (os->map_pos_y != 0) {
			if (os->map_pos_y < 0) {
				if (os->distance_pos_x <= 0) {
					os->displayed = 0;
					drawSprite = false;
				}
			} else {
				if (os->ref_ref_index < 160) {
					++os->ref_ref_index;
				}
				if (os->distance_pos_x >= 200) {
					os->displayed = 0;
					drawSprite = false;
				}
			}
		}
		if (drawSprite) {
			os->distance_pos_x += os->map_pos_y;
			os->distance_pos_y += g_game.cycle_counter;
			os->map_pos_x = (os->map_pos_x + angleInc) & 1023;
			const int angle = os->map_pos_x >> 1;
			os->pos_x = (int)(cos(angle) * os->distance_pos_x + .5) + dx;
			os->pos_y = (int)(sin(angle) * os->distance_pos_y + .5) + dy;
			g_game._render->addToSpriteList(g_game._res._oscar_spr, 20 + os->anim_data_cycles, os->pos_x, os->pos_y);
		}
	}
loc_1BF2:
	if (g_game.player_state.ram_duration >= 300 || inp_fire_button == 0) {
		if (g_game.player_state.ram_duration != 0 || --g_game.player_state.ram_duration != 0) {
			goto loc_1CC1;
		}
	}
	g_game.player_state.ram_duration = 0;
	if (g_game.boss_colliding_op_flag > 0) {
		int offset = (g_game.player_state.pos_x + g_game.player_state.pos_dx_2 + 16) >> 4;
		offset = offset * _screen_cdg_tile_map_h + g_game.decor_state.tile_map_offs_x;
		offset += (g_game.player_state.pos_y + g_game.player_state.pos_dy_2) >> 4;
		offset += g_game.decor_state.tile_map_offs_y;
		g_game.player_state.decor_ref_flag0 = 0;
		game_set_decor_tile_flag0(2, offset);
		if ((g_game.player_state.decor_ref_flag0 & 2) != 0) {
			goto loc_1BF2;
		}
		++offset;
		game_set_decor_tile_flag0(2, offset);
		if ((g_game.player_state.decor_ref_flag0 & 2) == 0) {
			g_game.player_state.ram_duration = 1;
			goto loc_1CC1;
		}
		g_game.boss_colliding_op_flag = -1;
		if ((g_game.player_state.hdir_flags & 1) == 0) {
			g_game.player_state.anim_frame_ptr = &carlos_anim_frames_table__v3[61];
		} else {
			g_game.player_state.anim_frame_ptr = &carlos_anim_frames_table__v3[45];
		}
		g_game.player_state.anim_frames_count = 2;
		game_handle_objects7_helper3();
		goto loc_1DE9;
	}
loc_1CC1:
	os = &g_game.objects_table_ptr7[0];
	if (g_game.boss_colliding_op_flag != 0 || inp_fire_button == 0 || g_game.boss_colliding_op_flag != 0 || os->displayed != 0) {
		game_handle_objects7_helper3();
	} else if (g_game.player_state.anim_frame_ptr == &carlos_anim_frames_table__v3[40] || g_game.player_state.anim_frame_ptr == &carlos_anim_frames_table__v3[37]) {
		int offset = (g_game.player_state.pos_x + g_game.player_state.pos_dx_2) >> 4;
		offset = offset * _screen_cdg_tile_map_h + g_game.decor_state.tile_map_offs_x;
		offset += (g_game.player_state.pos_y + g_game.player_state.pos_dy_2 + g_game.player_state.dim_h) >> 4;
		offset += g_game.decor_state.tile_map_offs_y;
		g_game.player_state.decor_ref_flag0 = 0;
		game_set_decor_tile_flag0(1, offset);
		if ((g_game.player_state.decor_ref_flag0 & 1) != 0) {
			offset += _screen_cdg_tile_map_h;
			game_set_decor_tile_flag0(1, offset);
			if ((g_game.player_state.decor_ref_flag0 & 1) != 0) {
				offset += _screen_cdg_tile_map_h;
				game_set_decor_tile_flag0(1, offset);
				if ((g_game.player_state.decor_ref_flag0 & 1) != 0) {
					if (g_game.player_state.rams_count != 0) {
						--g_game.player_state.rams_count;
						g_game.boss_colliding_op_flag = 1;
						g_game.player_state.hdir_flags &= ~0x40;
						if ((g_game.player_state.hdir_flags & 1) == 0) {
							g_game.player_state.anim_frame_ptr = &carlos_anim_frames_table__v3[67];
						} else {
							g_game.player_state.anim_frame_ptr = &carlos_anim_frames_table__v3[51];
						}
						g_game.player_state.anim_frames_count = 2;
						g_game.player_state.ram_duration = 375;
					}
					game_handle_objects7_helper3();
				}
			}
		}
	}
loc_1DE9:
	if (g_game.player_state.dead_flag == 0) {
		g_game.update_nicky_anim_flag = -1;
		os = &g_game.objects_table_ptr7[0];
		if (os->displayed != 0 && g_game.player_state.has_blue_ball == 0) {
			g_game.update_nicky_anim_flag = 0;
		}
	}
}

void game_set_next_carlos_anim_frame() {
	const anim_frame_t *af = g_game.player_state.anim_frame_ptr;
	if (af->frames_count != 0) {
		--g_game.player_state.anim_frames_count;
		if (g_game.player_state.anim_frames_count <= 0) {
			af = af->next_anim_frame;
			g_game.player_state.anim_frame_ptr = af;
			g_game.player_state.anim_frames_count = af->frames_count;
		}
	}
}

void Game::update_carlos_object() {
	if (g_game.carlos_direction != 0 && (inp_direction_mask & 3) != 0) {
		g_game.carlos_direction = 0;
	}
	if (g_game.nicky_anim_flags1 != 0) {
		inp_direction_mask = 8;
		if ((g_game.player_state.hdir_flags & 1) != 0) {
			inp_direction_mask |= 1;
		} else {
			inp_direction_mask |= 2;
		}
	}
	g_game.nicky_anim_flags1 = 0;
	g_game.player_state.decor_ref_flag2 = 0;
	g_game.player_state.decor_ref_flag0 = 0;
	g_game.decor_state.scroll_inc = 0;
	if (g_game.update_nicky_anim_flag != 0) {
		if (g_game.player_state.unk3D != 0) {
			inp_direction_mask &= ~0xF;
			inp_fire_button = 0;
			--g_game.player_state.unk3D;
			if (g_game.player_state.unk3D == 0) {
				g_game.player_state.hdir_flags &= 0x7F;
			}
		}
		if ((inp_direction_mask & 4) == 0) {
			g_game.player_state.vdir_flags |= 1;
		} else if (g_game.player_state.vdir_flags == 3) {
			g_game.player_state.vdir_flags = 0;
		}
		game_update_anim_helper();
		if ((g_game.player_state.hdir_flags & 8) != 0) {
			game_set_next_carlos_anim_frame();
			if (g_game.player_state.anim_frame_ptr->frames_count == 0) {
				g_game.player_state.hdir_flags &= ~8;
			}
		}
		if (g_game.player_state.shield_duration != 0) {
			--g_game.player_state.shield_duration;
		}
		if ((g_game.player_state.decor_ref_flag2) != 0) {
			const uint8 priority = SAMPLES_SND[0 * 16 + 12];
			if (SAMPLES_SND[0 * 16 + 12] > _snd_current_sfx_priority) {
				_snd_current_sfx_priority = priority;
				_snd_current_sfx_num = 0;
			}
		}
		return;
	}
	if (g_game.player_state.dead_flag != 0) {
		g_game.player_state.unk3D = 0;
		if (g_game.player_state.pos_y > 160) {
			if (g_game.quit_level_flag != 0) {
				g_game.quit_level_flag = 2;
			} else {
				g_game.restart_level_flag = 1;
			}
			return;
		}
		int dy = g_game.player_state.move_offsets_data_ptr[0];
		if (dy == 0x8888) {
			--g_game.player_state.move_offsets_data_ptr;
			dy = g_game.player_state.move_offsets_data_ptr[0];
		}
		++g_game.player_state.move_offsets_data_ptr;
		g_game.player_state.pos_y += (int16)dy;
		if ((g_game.player_state.hdir_flags & 1) != 0) {
			++g_game.player_state.pos_x;
		} else {
			--g_game.player_state.pos_x;
		}
		return;
	}
	if (g_game.player_state.action_code != 0) {
		if (g_game.player_state.action_code == 1) {
			if ((g_game.player_state.hdir_flags & 1) == 0) {
				g_game.player_state.anim_frame_ptr = &carlos_anim_frames_table__v3[43];
			} else {
				g_game.player_state.anim_frame_ptr = &carlos_anim_frames_table__v3[41];
			}
			g_game.player_state.anim_frames_count = 2;
			g_game.player_state.action_code = 2;
		}
		game_set_next_carlos_anim_frame();
	}
}

void Game::update_object_tilemap(object_state_t *os) {
	const int offset = (os->map_pos_x >> 4) * _screen_cdg_tile_map_h + (os->map_pos_y >> 4);
	if ((os->tile_num & 0x80) == 0) {
		const uint8 *src = g_game._res.get_new_ptr(os->tile_num);
		test_object_bounding_box_offset = offset;
		int sz = src[0];
		test_object_bounding_box_counter = src[1];
		src += 2;
		for (; sz != 0; --sz) {
			for (int i = 0; i < test_object_bounding_box_counter; ++i) {
				if (*src != 255) {
					g_game._res._decor_cdg[offset + i * _screen_cdg_tile_map_h] = *src;
				}
				++src;
			}
		}
		test_object_bounding_box_result = -1;
	}
}

Game g_game;

struct CarlosGameStub : GameStub {
	int _state;
	virtual int C init(int argc, char *argv[]) {
		fio_init(argc == 0 ? "DATA" : argv[1]);
		// TODO: check datafiles
		g_game.init();
		inp_init(IRS_NONE);
		_state = 0;
		return 0;
	}
	virtual void C quit() {
		g_game._render->destroy();
	}
	virtual void C save() {
		// TODO
	}
	virtual void C doTick(int keyMask) {
		inp_copy_state(keyMask);
		switch (_state) {
		case 0:
			game_init_level(g_game._currentLevel);
			_state = 1;
			break;
		}
		game_run_cycle();
		if (g_game.quit_level_flag >= 2) {
			_state = 0;
		} else if (g_game.new_level_flag) {
			inp_end_level();
			++g_game._currentLevel;
			if (g_game._currentLevel == 5) {
				g_game._currentLevel = 0;
			}
			_state = 0;
		}
	}
	virtual void C initGL(int w, int h) {
		if (!g_game._render) {
			g_game._render = new Render;
		}
		g_game._render->init(w, h);
	}
	virtual void C drawGL() {
		g_game._render->updateScreen();
		g_game._render->clearSpriteList();
		if (_state == 0) {
			g_game._render->flushTextures();
		}
	}
};

extern "C" {
	DYNLIB_SYMBOL GameStub *GameStub_create() {
		GameStub *s = new CarlosGameStub;
		return s;
	}
}

