/*
 * Carlos engine rewrite
 * Copyright (C) 2006-2011 Gregory Montoir (cyx@users.sourceforge.net)
 */

#ifndef GAME_H__
#define GAME_H__

#include "util.h"
#include "defs.h"
#include "resource.h"

struct Render;

#define GAME_SCREEN_W 320
#define GAME_SCREEN_H 160
#define GAME_SCROLL_W (GAME_SCREEN_W / 4)

#define GAME_ANIM_NICKY_DX 4
#define GAME_ANIM_NICKY_DY 4

#define GAME_DIR_RIGHT 1
#define GAME_DIR_LEFT  2
#define GAME_DIR_UP    4
#define GAME_DIR_DOWN  8
#define GAME_DIR_VERT  (GAME_DIR_UP | GAME_DIR_DOWN)
#define GAME_DIR_HORIZ (GAME_DIR_RIGHT | GAME_DIR_LEFT)
#define GAME_DIR_ANY   (GAME_DIR_RIGHT | GAME_DIR_LEFT | GAME_DIR_UP | GAME_DIR_DOWN)

#define GAME_INV_LIFES 0
#define GAME_INV_KEYS  1
#define GAME_INV_RAMS  2
#define GAME_INV_WINDS 3

void game_init_level(int level);
void game_init_carlos();
void game_update_life_bar();
void game_run_cycle();
void game_init_player_state();
void game_play_sound();
void game_draw_object(object_state_t *os);
void game_draw_carlos();
void game_redraw_inventory();
uint8 game_set_decor_tile_flags4_3_2_0(uint8 b, int offset);
void game_set_decor_tile_flag0(uint8 b, int offset);
uint8 game_get_decor_tile_flag3(int offset);
void game_set_next_carlos_anim_frame();
void game_reset_objects8();
void game_init_level_start_screen();
void game_init_ref_ref(anim_data_t *ad, uint16 flags);
void game_init_objects_from_positref(const uint8 *posit_ref, uint16 objects_offset);
void game_execute_logic_op();
void game_update_anim_helper();
void game_change_decor_tile_map(const object_state_t *os);
void game_update_score();
void game_adjust_tilemap();
void game_handle_level_change();
void game_execute_colliding_op();
void game_execute_colliding_op_helper1();
void game_execute_colliding_op_helper2();
void game_execute_colliding2_op();
void game_play_anim_data_sfx(const anim_data_t *ad);
void game_setup_monster(int flag);
anim_data_t *game_init_object_from_anim3(object_state_t *os, anim_data_t *ad);
void game_handle_projectiles_collisions_helper2(object_state_t *os, int _bp, int count);
anim_data_t *game_logic_op_helper1(object_state_t *os, anim_data_t *ad);
void game_logic_op_helper2(object_state_t *_di, anim_data_t *_si, anim_data_t *ad, uint16 move);
anim_data_t *game_logic_op2(object_state_t *os, anim_data_t *ad);
anim_data_t *game_logic_op(object_state_t *os, anim_data_t *ad);
void game_collides_op(object_state_t *os, anim_data_t *ad);
void game_collides_op_oscar(object_state_t *os, anim_data_t *ad);
int game_get_decor_offset(int x, int y);
int game_get_nicky_offset(int x, int y);
int game_test_object_bounding_box(object_state_t *os, anim_data_t *ad, int x1, int y1, int x2, int y2);
int game_test_object_on_screen(object_state_t *os, anim_data_t *ad);
void game_collide_bonus(int offset);
void game_update_objects_bbox();
void game_update_objects_bbox_helper1(object_state_t *os);
void game_update_objects_bbox_helper2(object_state_t *os);
void game_update_boss_object();
void game_handle_objects7();
void game_update_carlos_object();
uint16 game_get_decor_tile_flag7(int offset);
void game_update_object_tilemap(object_state_t *os);

#define OBJECTS_SIZE 349

struct Game {
	Render *_render;
	Resource _res;
	int _currentLevel;
	player_state_t player_state;
	oscar_state_t _oscarState;
	game_state_t game_state;
	decor_state_t decor_state;
	boss_state_t bossState;
	int player_pos_x, player_pos_y;
	int cycle_counter, cycle_counter2;
	int bonus_count;
	int cycle_score;
	object_state_t objects_table[OBJECTS_SIZE];
	object_state_t *objects_table_ptr1;
	object_state_t *objects_table_ptr2;
	object_state_t *objects_table_ptr7;
	object_state_t *objects_table_ptr8;
	object_state_t *objects_table_ptr9;
	int16 _oscarPosTable[40];
	anim_data_t *ref_ref_1, *ref_ref_21;
	int unk_ref_index;
	object_state_t *colliding_boss_object;
	int colliding_boss_object_flag;
	int boss_logic_op_flag, boss_colliding_op_flag;
	int winds_count;
	int carlos_direction;
	int carlos_offset_dy;
	int test_object_bounding_box_offset;
	int test_object_bounding_box_counter;
	int test_object_bounding_box_result;
	int execute_op_logic_helper1_flag;
	int monster_fight_flag;
	int restart_level_flag;
	int new_level_flag;
	int quit_level_flag;
	int op_colliding_mask;
	uint16 nicky_anim_flags1;
	int update_nicky_anim_flag;
	int update_decor_counter;
	const uint16 *carlos_move_offsets_ptr;
	int bounding_box_x1, bounding_box_x2, bounding_box_y1, bounding_box_y2;
	object_state_t *objects_list_cur, *objects_list_head, *objects_list_last;
	int ret_flag;
	int life_bar_flag;

	void init();
	void init_player_state();
	void init_oscar_state();
	void draw_oscar_object();
	void update_oscar_object();
	void fix_player_pos_tilemap_offset();
	void update_carlos_object();
	void update_object_tilemap(object_state_t *os);
};

#endif // GAME_H__
