/*
 * Carlos engine rewrite
 * Copyright (C) 2006-2011 Gregory Montoir (cyx@users.sourceforge.net)
 */

#include "fileio.h"
#include "game.h"
#include "globals.h"
#include "input.h"
#include "stub.h"

static int file_slot;
static int cur_key_mask;
static const uint8 *inp_buf_ptr, *inp_buf_end;
static input_recording_state_e recording_state;

static void inp_record(int key_mask) {
	if (file_slot >= 0) {
		uint8 b = key_mask & (SKM_UP | SKM_DOWN | SKM_RIGHT | SKM_LEFT | SKM_ACTION);
		fio_write(file_slot, &b, 1);
	}
}

static int inp_replay() {
	int key_mask = 0;
	if (inp_buf_ptr < inp_buf_end) {
		uint8 b = *inp_buf_ptr++;
		key_mask = b;
	}
	return key_mask;
}

void inp_init(input_recording_state_e rec_state) {
	file_slot = -1;
	recording_state = rec_state;
}

void inp_start_level(int level_num) {
	char filename[32];
	snprintf(filename, sizeof(filename), "CARLOS-L%d.BIN", level_num + 1);
	switch (recording_state) {
	case IRS_REPLAY:
		file_slot = fio_open(filename, FIO_READ, 0);
		if (file_slot >= 0) {
			inp_buf_ptr = fio_fetch(file_slot);
			inp_buf_end = inp_buf_ptr + fio_size(file_slot);
			util_random_seed = READ_BE_UINT16(inp_buf_ptr); inp_buf_ptr += 2;
			fio_close(file_slot);
			file_slot = -1;
		} else {
			inp_buf_ptr = 0;
			inp_buf_end = 0;
		}
		break;
	case IRS_RECORD:
		file_slot = fio_open(filename, FIO_WRITE, 0);
		if (file_slot >= 0) {
			uint8 tmp[2];
			tmp[0] = util_random_seed >> 8;
			tmp[1] = util_random_seed & 0xFF;
			fio_write(file_slot, tmp, 2);
		}
		break;
	case IRS_NONE:
		break;
	}
}

void inp_end_level() {
	if (file_slot >= 0) {
		fio_close(file_slot);
		file_slot = -1;
	}
}

static int inp_is_key_pressed(int prev_key_mask, int mask) {
	if ((prev_key_mask & mask) != 0 && (cur_key_mask & mask) == 0) {
		/* key was down and is now up */
		return 1;
	}
	return 0;
}

static void inp_read_state(int key_mask) {
	const int prev_key_mask = cur_key_mask;
	cur_key_mask = key_mask;
	switch (recording_state) {
	case IRS_REPLAY:
		cur_key_mask |= inp_replay();
		inp_key_pressed_action = (cur_key_mask & SKM_ACTION) != 0;
		break;
	case IRS_RECORD:
		inp_record(cur_key_mask);
		break;
	case IRS_NONE:
		break;
	}
	inp_key_pressed_action = inp_is_key_pressed(prev_key_mask, SKM_ACTION);
	inp_key_pressed_quit = inp_is_key_pressed(prev_key_mask, SKM_QUIT);
	inp_key_pressed_f[1] = inp_is_key_pressed(prev_key_mask, SKM_F1);
	inp_key_pressed_f[2] = inp_is_key_pressed(prev_key_mask, SKM_F2);
	inp_key_pressed_f[3] = inp_is_key_pressed(prev_key_mask, SKM_F3);
	inp_key_pressed_f[4] = inp_is_key_pressed(prev_key_mask, SKM_F4);
}

void inp_copy_state(int key_mask) {
	inp_read_state(key_mask);
	/* handle special keys */
	if (inp_key_pressed_quit) {
		inp_key_pressed_quit = 0;
		g_game.quit_level_flag = 2;
	}
	if (inp_key_pressed_f[1]) {
		inp_key_pressed_f[1] = 0;
		g_game.player_state.shield_duration = 1900;
		g_game.player_state.lifes_count = 9;
		g_game.player_state.keys_count = 9;
		g_game.player_state.rams_count = 9;
		g_game.player_state.winds_count = 9;
		info("Cheat mode, adding all bonuses");
	}
	if (inp_key_pressed_f[2]) {
		inp_key_pressed_f[2] = 0;
		g_game.new_level_flag = 1;
		info("Cheat mode, jumping to next level");
	}
	if (inp_key_pressed_f[4]) {
		inp_key_pressed_f[4] = 0;
		g_game.player_state.shield_duration = 0;
		g_game.player_state.fill_energy_bar = 0;
		g_game.player_state.colliding = 128;
		info("Cheat mode, committing suicide");
	}
	if (inp_key_pressed_f[3]) {
		inp_key_pressed_f[3] = 0;
		const int x = _screen_cdg_tile_map_w - (GAME_SCREEN_W >> 4);
		const int y = _screen_cdg_tile_map_h - (GAME_SCREEN_H >> 4);
		g_game.player_state.tilemap_offset = x * _screen_cdg_tile_map_h + y;
		g_game.fix_player_pos_tilemap_offset();
		g_game.restart_level_flag = 1;
		info("Cheat mode, jumping to boss");
	}
	/* handle game keys */
	inp_direction_mask = 0;
	if ((cur_key_mask & (SKM_UP | SKM_DOWN)) != (SKM_UP | SKM_DOWN)) {
		if (cur_key_mask & SKM_UP) {
			inp_direction_mask |= GAME_DIR_UP;
		}
		if (cur_key_mask & SKM_DOWN) {
			inp_direction_mask |= GAME_DIR_DOWN;
		}
	}
	if ((cur_key_mask & (SKM_LEFT | SKM_RIGHT)) != (SKM_LEFT | SKM_RIGHT)) {
		if (cur_key_mask & SKM_LEFT) {
			inp_direction_mask |= GAME_DIR_LEFT;
		}
		if (cur_key_mask & SKM_RIGHT) {
			inp_direction_mask |= GAME_DIR_RIGHT;
		}
	}
	inp_fire_button = (cur_key_mask & SKM_ACTION) != 0;
}
