/*
 * Carlos engine rewrite
 * Copyright (C) 2006-2011 Gregory Montoir (cyx@users.sourceforge.net)
 */

#include <SDL.h>
#include <SDL_opengl.h>
#ifdef _WIN32
#include <windows.h>
#endif
#include "stub.h"

static int gWindowW = 320 * 3;
static int gWindowH = 160 * 3;
static const int gTickDuration = 2000 / 60;
static const char *gWindowTitle = "Carlos";

#ifdef _WIN32
static const char *gSoSym = "GameStub_create";
static const char *gSoName = "carlos.dll";
struct GetStub_impl {
	HINSTANCE _dlso;

	GetStub_impl()
		: _dlso(0) {
	}
	~GetStub_impl() {
		if (_dlso) {
			FreeLibrary(_dlso);
			_dlso = 0;
		}
	}
	void *open(const char *name) {
		_dlso = LoadLibrary(name);
		return _dlso;
	}
	void *getSymbol(const char *name) {
		return (void *)GetProcAddress(_dlso, name);
	}
	GameStub *getGameStub() {
		if (open(gSoName)) {
			void *proc = getSymbol(gSoSym);
			if (proc) {
				typedef GameStub *(*createProc)();
				return ((createProc)proc)();
			}
		}
		return 0;
	}
};
#else
extern "C" {
	GameStub *GameStub_create();
}
struct GetStub_impl {
	GameStub *getGameStub() {
		return GameStub_create();
	}
};
#endif

int main(int argc, char *argv[]) {
	GetStub_impl gs;
	GameStub *stub = gs.getGameStub();
	if (!stub) {
		return 0;
	}
	int ret = stub->init(argc - 1, argv + 1);
	if (ret != 0) {
		return ret;
	}
	SDL_Init(SDL_INIT_VIDEO);
	SDL_ShowCursor(SDL_DISABLE);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	SDL_SetVideoMode(gWindowW, gWindowH, 0, SDL_OPENGL | SDL_RESIZABLE);
	SDL_WM_SetCaption(gWindowTitle, 0);
	stub->initGL(gWindowW, gWindowH);
	bool quitGame = false;
	bool paused = false;
	int _keyMask = 0;
	while (!quitGame) {
		SDL_Event ev;
		while (SDL_PollEvent(&ev)) {
			switch (ev.type) {
			case SDL_QUIT:
				quitGame = true;
				break;
			case SDL_ACTIVEEVENT:
				if (ev.active.state & SDL_APPINPUTFOCUS) {
					paused = !ev.active.gain;
				}
				break;
			case SDL_VIDEORESIZE:
				gWindowW = ev.resize.w;
				gWindowH = ev.resize.h;
				SDL_SetVideoMode(gWindowW, gWindowH, 0, SDL_OPENGL | SDL_RESIZABLE);
				stub->initGL(gWindowW, gWindowH);
				break;
			case SDL_KEYUP:
				switch (ev.key.keysym.sym) {
				case SDLK_LEFT:
					_keyMask &= ~SKM_LEFT;
					break;
				case SDLK_RIGHT:
					_keyMask &= ~SKM_RIGHT;
					break;
				case SDLK_UP:
					_keyMask &= ~SKM_UP;
					break;
				case SDLK_DOWN:
					_keyMask &= ~SKM_DOWN;
					break;
				case SDLK_RCTRL:
				case SDLK_SPACE:
					_keyMask &= ~SKM_ACTION;
					break;
				case SDLK_F1:
					_keyMask &= ~SKM_F1;
					break;
				case SDLK_F2:
					_keyMask &= ~SKM_F2;
					break;
				case SDLK_F3:
					_keyMask &= ~SKM_F3;
					break;
				case SDLK_F4:
					_keyMask &= ~SKM_F4;
					break;
				case SDLK_ESCAPE:
					_keyMask &= ~SKM_QUIT;
					break;
				default:
					break;
				}
				break;
			case SDL_KEYDOWN:
				switch (ev.key.keysym.sym) {
				case SDLK_LEFT:
					_keyMask |= SKM_LEFT;
					break;
				case SDLK_RIGHT:
					_keyMask |= SKM_RIGHT;
					break;
				case SDLK_UP:
					_keyMask |= SKM_UP;
					break;
				case SDLK_DOWN:
					_keyMask |= SKM_DOWN;
					break;
				case SDLK_RCTRL:
				case SDLK_SPACE:
					_keyMask |= SKM_ACTION;
					break;
				case SDLK_F1:
					_keyMask |= SKM_F1;
					break;
				case SDLK_F2:
					_keyMask |= SKM_F2;
					break;
				case SDLK_F3:
					_keyMask |= SKM_F3;
					break;
				case SDLK_F4:
					_keyMask |= SKM_F4;
					break;
				case SDLK_ESCAPE:
					_keyMask |= SKM_QUIT;
					break;
				default:
					break;
				}
				break;
			}
		}
		if (!paused) {
			stub->doTick(_keyMask);
			stub->drawGL();
			SDL_GL_SwapBuffers();
		}
		SDL_Delay(gTickDuration);
	}
	stub->save();
	stub->quit();
	return 0;
}

