/*
 * Carlos engine rewrite
 * Copyright (C) 2006-2011 Gregory Montoir (cyx@users.sourceforge.net)
 */

#include "game.h"
#include "globals.h"
#include "render.h"
#include "sound.h"

static void game_collides_play_sound(object_state_t *os) {
	anim_data_t *ad = os->anim_data1_ptr;
	os->displayed = 0;
	if (ad->sound_num >= 0) {
		snd_play_sfx(&SAMPLES_SND[ad->sound_num * 16]);
	}
}

static void game_duplicate_object_play_sfx(object_state_t *os, anim_data_t *ad) {
	os->anim_data2_ptr = ad->anim_data2_ptr;
	if (ad->anim_data3_ptr == 0) {
		os->displayed = 0;
	} else {
		ad = ad->anim_data3_ptr;
		os->anim_data_cycles = ad->cycles;
		os->unk3 = ad->unk3;
		os->unk26 = ad->unk4;
		os->unk27 = ad->unk5;
		os->move_data_ptr = &ad->move_data_ptr[ad->init_sprite_num];
		os->distance_pos_x = ad->distance_dx;
		os->distance_pos_y = ad->distance_dy;
		os->anim_data4_ptr = ad->anim_data2_ptr;
		os->anim_data1_ptr = ad;
		os->anim_data3_ptr = ad;
		game_play_anim_data_sfx(ad);
	}
}

static void game_remove_object_play_sfx(object_state_t *os, anim_data_t *ad) { /* _di, _bx */
	if (ad->anim_data3_ptr == 0) {
		os->displayed = 0;
		if (os->visible == 0) {
			return;
		}
		++os->displayed;
		game_play_anim_data_sfx(os->anim_data1_ptr);
	} else {
		if (os->visible == 0) {
			os->anim_data2_ptr = 0;
		}
		game_play_anim_data_sfx(game_init_object_from_anim3(os, ad));
	}
}

static void game_colliding_op27(object_state_t *os, anim_data_t *ad) {
	++os->transparent_flag;
	if (g_game.boss_colliding_op_flag == 0) {
		++g_game.player_state.colliding;
	}
	if (g_game.monster_fight_flag) {
		++g_game.player_state.colliding;
	}
}

static void game_colliding_op0_helper(object_state_t *os, anim_data_t *ad) {
	if (os->life > 0) {
		g_game.player_state.colliding_flag = 1;
	} else {
		os->transparent_flag = 0;
		g_game.cycle_score += ad->score;
		g_game.player_state.colliding_flag = 1;
		for (int i = 0; i < 5; ++i) {
			object_state_t *_si = &g_game.objects_table_ptr9[i];
			if (_si->displayed == 0) {
				++g_game.bonus_count;
				if (g_game.unk_ref_index >= 10) {
					g_game.unk_ref_index = 0;
				}
				++g_game.unk_ref_index;
				anim_data_t *_dx = &g_game.ref_ref_1[g_game.unk_ref_index - 1];
				_si->anim_data1_ptr = _si->anim_data3_ptr = _dx;
				_si->displayed = _dx->unk0;
				_si->life = _dx->lifes;
				_si->anim_data_cycles = _dx->cycles;
				_si->unk3 = _dx->unk3;
				_si->displayed = 2;
				_si->anim_data4_ptr = _dx->anim_data2_ptr;
				_si->move_data_ptr = _dx->move_data_ptr;
				_si->unk26 = _dx->unk4;
				_si->unk27 = _dx->unk5;
				_si->pos_x = os->pos_x;
				_si->pos_y = os->pos_y;
				_si->pos_x += (ad->anim_w >> 1) - (_dx->anim_w >> 1);
				_si->distance_pos_x = 0;
				_si->distance_pos_y = 0;
				if (ad->colliding_opcode != 24) {
					break;
				}
				/* make 2 monsters out of it */
				_dx = ad->anim_data4_ptr;
				for (int j = 0; j < 2; ++j) {
					_si = &os[1 + j];
					_si->anim_data1_ptr = _dx;
					_si->anim_data3_ptr = _dx;
					_si->displayed = _dx->unk0;
					_si->life = _dx->lifes;
					_si->anim_data_cycles = _dx->cycles;
					_si->unk3 = _dx->unk3;
					_si->displayed = 2;
					_si->anim_data4_ptr = _dx->anim_data2_ptr;
					_si->move_data_ptr = _dx->move_data_ptr;
					_si->distance_pos_x = _dx->distance_dx;
					_si->distance_pos_y = _dx->distance_dy;
					_si->unk26 = _dx->unk4;
					_si->unk27 = _dx->unk5;
					_si->pos_x = os->pos_x + ad->dx2;
					_si->pos_y = os->pos_y + ad->dy2 + ad->anim_h - _dx->anim_h;
					_si->map_pos_y -= _dx->anim_h - 1;
					_si->anim_data2_ptr = _dx->anim_data2_ptr;
					++_dx;
				}
				break;
			}
		}
		game_remove_object_play_sfx(os, ad);
	}
}

static void game_colliding_op0(object_state_t *os, anim_data_t *ad) {
	if (ad->lifes == 0 || (g_game.player_state.hdir_flags & 4) == 0) {
		game_colliding_op27(os, ad);
		return;
	}
	int y1 = g_game.player_state.dim_h / 2 + g_game.player_state.pos_y;
	int y2 = os->pos_y + ad->bounding_box_y1;
	if (y1 > y2) {
		game_colliding_op27(os, ad);
		return;
	}
	os->transparent_flag |= 1;
	os->life -= 8;
	game_colliding_op0_helper(os, ad);
}

static void game_colliding2_op0(object_state_t *os, anim_data_t *ad) {
	++os->transparent_flag;
	if (g_game.monster_fight_flag == 0) {
		os->life -= 2;
	} else {
		os->life -= 8;
	}
	game_colliding_op0_helper(os, ad);
}

static void game_colliding_op1(object_state_t *os, anim_data_t *ad) {
	g_game.player_state.fill_energy_bar = 1;
	g_game.player_state.colliding = 0;
	game_collides_play_sound(os);
}

static void game_colliding_op2(object_state_t *os, anim_data_t *ad) {
	if (g_game.player_state.lifes_count < 9) {
		++g_game.player_state.lifes_count;
	}
	game_collides_play_sound(os);
}

static void game_colliding_op3(object_state_t *os, anim_data_t *ad) {
	if (g_game.player_state.keys_count < 9) {
		++g_game.player_state.keys_count;
	}
	game_collides_play_sound(os);
}

static void game_colliding_op16(object_state_t *os, anim_data_t *ad) {
	object_state_t *_si = os + 1;
	_si->anim_data2_ptr = &g_game.ref_ref_21[os->ref_ref_index];
	_si->displayed = _si->anim_data2_ptr->unk0;
	_si->map_pos_y += ad->dy2;
	_si->map_pos_x += ad->dx2;
	os->anim_data2_ptr = ad->anim_data3_ptr;
	game_remove_object_play_sfx(os, ad);
}

static void game_colliding_op6(object_state_t *os, anim_data_t *ad) {
	g_game.player_state.shield_duration = 100;
	game_collides_play_sound(os);
}

static void game_colliding_op7(object_state_t *os, anim_data_t *ad) {
	g_game.player_state.tile_blinking_duration = 375;
	game_collides_play_sound(os);
}

static void game_colliding_op5(object_state_t *os, anim_data_t *ad) {
	if (g_game.player_state.rams_count < 9) {
		++g_game.player_state.rams_count;
	}
	game_collides_play_sound(os);
}

static void game_colliding_op8(object_state_t *os, anim_data_t *ad) {
	if (g_game.player_state.winds_count < 9) {
		++g_game.player_state.winds_count;
	}
	game_collides_play_sound(os);
}

static void game_colliding_op9(object_state_t *os, anim_data_t *ad) {
	++g_game.winds_count;
	if (g_game.winds_count == 10) {
		g_game.winds_count = 0;
		game_colliding_op8(os, ad);
	} else {
		game_collides_play_sound(os);
	}
}

static void game_colliding_op13_helper(object_state_t *os, anim_data_t *ad) {
	if (g_game.boss_colliding_op_flag != 0) {
		return;
	}
//	byte_1844E = 0;
	g_game.player_state.tilemap_offset = os->ref_ref_index - _screen_cdg_tile_map_h - 1;
	g_game.player_state.unk1 = 0;
	g_game.player_state.hdir_flags &= ~4;
//	word_183D9 = os->tile_num & 0x80;
//	screen_fade_in_pal();
	g_game.fix_player_pos_tilemap_offset();
	g_game._res.load_fond(os->tile_num & 0x7F, g_game._render);
	g_game.restart_level_flag = 1;
	uint8 displayed = os->displayed;
	game_collides_play_sound(os);
	os->displayed = displayed;
}

static void game_colliding_op11(object_state_t *os, anim_data_t *ad) {
	if ((inp_direction_mask & GAME_DIR_DOWN) == 0) {
		return;
	}
	game_colliding_op13_helper(os, ad);
}

static void game_colliding_op13(object_state_t *os, anim_data_t *ad) {
	if ((inp_direction_mask & GAME_DIR_UP) == 0) {
		return;
	}
	game_colliding_op13_helper(os, ad);
}

static void game_colliding_op12(object_state_t *os, anim_data_t *ad) {
	if (g_game.op_colliding_mask == 15) {
		game_colliding_op13(os, ad);
	}
}

static void game_colliding_op14(object_state_t *os, anim_data_t *ad) {
	os->anim_data2_ptr = 0;
	g_game.cycle_score += ad->score;
	game_remove_object_play_sfx(os, ad);
}

static void game_colliding_op15(object_state_t *os, anim_data_t *ad) {
	object_state_t *os2 = g_game.objects_table_ptr8;
	anim_data_t *ad2 = ad->anim_data4_ptr;
	os->anim_data2_ptr = ad->anim_data3_ptr;
	for (int i = 0; i < 5; ++i) {
		os2->anim_data1_ptr = ad2;
		os2->anim_data3_ptr = ad2;
		os2->displayed = ad2->unk0;
		os2->life = ad2->lifes;
		os2->anim_data4_ptr = ad2->anim_data2_ptr;
		os2->move_data_ptr = ad2->move_data_ptr + 2 * getRandomNumber(2);
		os2->anim_data_cycles = ad2->cycles;
		os2->unk3 = ad2->unk3;
		os2->unk26 = ad2->unk4;
		os2->unk27 = ad2->unk5;
		os2->pos_x = os->pos_x + ad->dx2 - (4 - i);
		os2->pos_y = os->pos_y + ad->dy2;
		os2->distance_pos_y = 0;
		os2->distance_pos_x = i - 2;
		++os2;
		++ad2;
	}
	++g_game.bonus_count;
	game_remove_object_play_sfx(os, ad);
}

static void game_colliding_op17(object_state_t *os, anim_data_t *ad) {
	os->anim_data2_ptr = ad->anim_data3_ptr;
	anim_data_t *ad2 = ad->anim_data4_ptr; // _dx
	for (int i = 0; i < 6; ++i) {
		object_state_t *_si = &g_game.objects_table_ptr8[i];
		_si->anim_data1_ptr = ad2;
		_si->anim_data3_ptr = ad2;
		_si->unk2C = 0;
		_si->displayed = ad2->unk0;
		_si->life = ad2->lifes;
		_si->anim_data_cycles = ad2->cycles;
		_si->unk3 = ad2->unk3;
		_si->anim_data4_ptr = ad2->anim_data2_ptr;
		_si->move_data_ptr = ad2->move_data_ptr;
//		_si->move_data_seg = getRandomNumber(2) * 2 + ad2->move_data_seg;
		_si->unk26 = ad2->unk4;
		_si->unk27 = ad2->unk5;
		_si->pos_x = os->pos_x + ad->dx2 - (5 - i);
		_si->pos_y = os->pos_y + ad->dy2;
		_si->distance_pos_y = 0;
		_si->distance_pos_x = getRandomNumber(9) - 4;
		++ad2;
	}
	++g_game.bonus_count;
	game_remove_object_play_sfx(os, ad);
}

static void game_colliding_op18(object_state_t *os, anim_data_t *ad) {
	os->anim_data1_ptr = ad->anim_data2_ptr;
}

static void game_colliding_op19(object_state_t *os, anim_data_t *ad) {
	if (g_game.boss_colliding_op_flag == 0 || g_game.player_state.keys_count != 0) {
		if (g_game.player_state.keys_count == 0) {
			return;
		}
		--g_game.player_state.keys_count;
	}
	anim_data_t *ad2 = os->anim_data1_ptr;
	if (ad2->logic_opcode >= 21 && ad2->logic_opcode <= 24) {
		g_game.update_object_tilemap(os);
	}
	game_duplicate_object_play_sfx(os, ad);
}

static void game_patch_player_move_offsets(int count) {
	static int16 move_offsets_tmp[12] = {
		0x7777, 0x7777, 0x7777, 0x7777, 0x7777, 0x7777,
		0x7777, 0x7777, 0x7777, 0x7777, 0x7777, 0x7777
	};
	g_game.player_state.move_offsets_data_ptr = (const uint16 *)move_offsets_tmp;
	int offset = 0;
	int i = 0;
	do {
		move_offsets_tmp[offset] = (offset & 1) == 0 ? -5 : -6;
		++offset;
		i += 5;
	} while (i < count * 16);
	for (i = 0;; ++i) {
		move_offsets_tmp[offset] = move_data[6 + i];
		++offset;
		if (move_data[6 + i] == 0x7777) {
			break;
		}
	}
}

static void game_colliding_op20(object_state_t *os, anim_data_t *ad) {
	int offset = (os->map_pos_x >> 4) * _screen_cdg_tile_map_h + (os->map_pos_y >> 4);
	if (os->tile_num != g_game._res._decor_cdg[offset] && os->tile_num != 255) {
		g_game._res._decor_cdg[offset] = os->tile_num;
		g_game.test_object_bounding_box_result = -1;
		game_play_anim_data_sfx(ad);
	}
	os->displayed = 0;
	os = &g_game.objects_list_head[os->ref_ref_index];
	ad = os->anim_data1_ptr;
	if (ad->logic_opcode >= 21 && ad->logic_opcode <= 24) {
		g_game.update_object_tilemap(os);
	}
	game_duplicate_object_play_sfx(os, ad);
}

static void game_colliding_op21(object_state_t *os, anim_data_t *ad) {
	if (g_game.player_state.hdir_flags & 4) {
		int y = os->pos_y + ad->bounding_box_y1;
		if (g_game.player_state.pos_y + g_game.player_state.dim_h / 2 <= y) {
			y -= g_game.player_state.dim_h;
			g_game.player_state.pos_y = y;
			g_game.player_state.unk1 = 2;
			g_game.player_state.hdir_flags &= ~4;
			game_patch_player_move_offsets(os->ref_ref_index);
			if (ad->colliding_opcode != 21) {
				g_game.player_state.move_offsets_data_ptr = (const uint16 *)&move_data[0];
			}
			game_remove_object_play_sfx(os, ad);
		}
	}
}

static void game_colliding_op22(object_state_t *os, anim_data_t *ad) {
	int offset = (os->map_pos_x >> 4) * _screen_cdg_tile_map_h + (os->map_pos_y >> 4);
	if (os->tile_num != g_game._res._decor_cdg[offset]) {
		g_game._res._decor_cdg[offset] = os->tile_num;
		g_game.test_object_bounding_box_result = -1;
		game_play_anim_data_sfx(ad);
	}
	os->displayed = 0;
	os = &g_game.objects_list_head[os->ref_ref_index];
	ad = os->anim_data1_ptr;
	os->anim_data2_ptr = ad->anim_data3_ptr;
	anim_data_t *ad2 = ad->anim_data4_ptr;
	for (int i = 0; i < 10; ++i) {
		object_state_t *_si = &g_game.objects_table_ptr8[i];
		_si->anim_data1_ptr = ad2;
		_si->anim_data3_ptr = ad2;
		_si->unk2C = 0;
		_si->displayed = ad2->unk0;
		_si->life = ad2->lifes;
		_si->anim_data_cycles = ad2->cycles;
		_si->unk3 = ad2->unk3;
		_si->unk26 = ad2->unk4;
		_si->unk27 = ad2->unk5;
		_si->move_data_ptr = ad2->move_data_ptr;
		_si->anim_data4_ptr = ad->anim_data2_ptr;
//		_si->move_data_seg = getRandomNumber(2) * 2 + ad2->move_data_seg;
		_si->pos_y = os->pos_y + ad->dy2;
		_si->pos_x = os->pos_x + ad->dx2 - (9 - i);
		_si->distance_pos_y = 0;
		_si->distance_pos_x = i - 2;
		++ad2;
	}
	++g_game.bonus_count;
	if (ad->anim_data3_ptr == 0) {
		os->displayed = 0;
		if (os->visible != 0) {
			++os->displayed;
		}
	} else {
		if (os->visible == 0) {
			os->anim_data2_ptr = 0;
		}
		game_init_object_from_anim3(os, ad);
	}
}

static void game_colliding_op23(object_state_t *os, anim_data_t *ad) {
	int offset = (os->map_pos_x >> 4) * _screen_cdg_tile_map_h + (os->map_pos_y >> 4);
	if (os->tile_num != g_game._res._decor_cdg[offset]) {
		g_game._res._decor_cdg[offset] = os->tile_num;
		g_game.test_object_bounding_box_result = -1;
		game_play_anim_data_sfx(ad);
	}
	os->displayed = 0;
	os = &g_game.objects_list_head[os->ref_ref_index];
	ad = os->anim_data1_ptr;
	anim_data_t *ad2 = &g_game.ref_ref_21[os->ref_ref_index];
	os[1].anim_data2_ptr = ad2;
	os[1].displayed = ad2->unk0;
	os[1].map_pos_y += ad->dy2;
	os[1].map_pos_x += ad->dx2;
	os->anim_data2_ptr = ad->anim_data3_ptr;
	if (ad->anim_data3_ptr == 0) {
		os->displayed = 0;
		if (os->visible == 0) {
			return;
		}
		++os->displayed;
	} else {
		game_init_object_from_anim3(os, ad);
	}
}

static void game_colliding_op28(object_state_t *os, anim_data_t *ad) {
	int offset = (os->map_pos_x >> 4) * _screen_cdg_tile_map_h + (os->map_pos_y >> 4);
	if (os->tile_num != g_game._res._decor_cdg[offset]) {
		g_game._res._decor_cdg[offset] = os->tile_num;
		g_game.test_object_bounding_box_result = -1;
		game_play_anim_data_sfx(ad);
	}
	os->displayed = 0;
	os = &g_game.objects_list_head[os->ref_ref_index];
	if (os->displayed != 0 && (os->displayed & 0x80) == 0) {
		game_duplicate_object_play_sfx(os, os->anim_data1_ptr);
	}
}

static void game_colliding_op25(object_state_t *os, anim_data_t *ad) {
	g_game.new_level_flag = 1;
	ad = os->anim_data1_ptr;
	if (ad->sound_num >= 0) {
		snd_play_sfx(&SAMPLES_SND[ad->sound_num * 16]);
	}
}

static void game_colliding_op26(object_state_t *os, anim_data_t *ad) {
	if (g_game.player_state.hdir_flags & 4) {
		int y = os->pos_y + ad->bounding_box_y1;
		if (g_game.player_state.pos_y + g_game.player_state.dim_h / 2 <= y) {
			y -= g_game.player_state.dim_h;
			g_game.player_state.pos_y = y;
			g_game.player_state.unk1 = 2;
			g_game.player_state.hdir_flags &= ~4;
			game_patch_player_move_offsets(os->ref_ref_index);
			game_remove_object_play_sfx(os, ad);
		}
	}
}

static void game_colliding_op29(object_state_t *os, anim_data_t *ad) {
	if ((os->tile_num & 0x80) == 0) {
		int offset = os->ref_ref_index;
		const uint8 *src = g_game._res.get_new_ptr(offset);
		g_game.test_object_bounding_box_offset = offset;
		int sz = src[0];
		g_game.test_object_bounding_box_counter = src[1];
		src += 2;
		for (; sz != 0; --sz) {
			for (int i = 0; i < g_game.test_object_bounding_box_counter; ++i) {
				g_game._res._decor_cdg[offset + i * _screen_cdg_tile_map_h] = *src;
				++src;
			}
		}
		g_game.test_object_bounding_box_result = -1;
		snd_play_sfx(&SAMPLES_SND[0 * 16]);
	}
}

static void game_colliding_op4(object_state_t *os, anim_data_t *ad) {
	g_game.carlos_move_offsets_ptr = (const uint16 *)&move_data[0];
	g_game.player_state.has_spring = 375;
	game_collides_play_sound(os);
}

static void game_colliding_op35_helper(object_state_t *os, anim_data_t *ad, int num) {
	g_game.op_colliding_mask |= 1 << num;
	static const uint8 coords[] = { 5, 0x50, 5, 0x60, 0x15, 0x50, 0x15, 0x60 };
	const int x = coords[num * 2];
	const int y = coords[num * 2 + 1];
	g_game._render->addToSpriteList(g_game._res._level_spr, g_game._res._decor_cdg[1 + num], x, y);
	game_collides_play_sound(os);
}

static void game_colliding_op32(object_state_t *os, anim_data_t *ad) {
	game_colliding_op35_helper(os, ad, 0);
}

static void game_colliding_op33(object_state_t *os, anim_data_t *ad) {
	game_colliding_op35_helper(os, ad, 1);
}

static void game_colliding_op34(object_state_t *os, anim_data_t *ad) {
	game_colliding_op35_helper(os, ad, 2);
}

static void game_colliding_op35(object_state_t *os, anim_data_t *ad) {
	game_colliding_op35_helper(os, ad, 3);
}

static void game_colliding_op41_helper(object_state_t *os, anim_data_t *ad, int num) {
	if ((os->ref_ref_index == 0 || g_game.boss_colliding_op_flag != 0) && g_game.cycle_counter2 != 0) {
		if ((os->tile_num & 0x80) == 0) {
			return;
		}
		const uint8 *src = g_game._res.get_new_ptr(os->tile_num) + 2; // _si
		int offset = (os->map_pos_x >> 4) * _screen_cdg_tile_map_h + (os->map_pos_y >> 4);
		offset -= num;
		++offset;
		int r = getRandomNumber(num);
		int i = 0;
		while (1) {
			if (src[r] != g_game._res._decor_cdg[offset + r]) {
				break;
			}
			++i;
			if (num == i) {
				break;
			}
			++r;
			if (r == num) {
				r = 0;
			}
		}
		g_game._res._decor_cdg[offset + r] = src[r];
		if (num == i) {
			os->displayed = 0;
			snd_play_sfx(&SAMPLES_SND[15 * 16]);
			g_game.test_object_bounding_box_result = -1;
		}
	}
}

static void game_colliding_op38(object_state_t *os, anim_data_t *ad) {
	game_colliding_op41_helper(os, ad, 5);
}

static void game_colliding_op39(object_state_t *os, anim_data_t *ad) {
	game_colliding_op41_helper(os, ad, 2);
}

static void game_colliding_op40(object_state_t *os, anim_data_t *ad) {
	game_colliding_op41_helper(os, ad, 3);
}

static void game_colliding_op41(object_state_t *os, anim_data_t *ad) {
	game_colliding_op41_helper(os, ad, 4);
}

static void game_colliding_op42(object_state_t *os, anim_data_t *ad) {
	if (g_game.boss_colliding_op_flag == 0) {
		g_game.player_state.shield_duration = 0;
		while (1) {
			g_game.player_state.colliding = 100;
			if (g_game.player_state.energy == 0) {
				break;
			}
			game_update_life_bar();
		}
	}
}

void game_collides_op(object_state_t *os, anim_data_t *ad) {
	switch (ad->colliding_opcode) {
	case 0:
		game_colliding_op0(os, ad);
		break;
	case 1:
		game_colliding_op1(os, ad);
		break;
	case 2:
		game_colliding_op2(os, ad);
		break;
	case 3:
		game_colliding_op3(os, ad);
		break;
	case 4:
		game_colliding_op4(os, ad);
		break;
	case 5:
		game_colliding_op5(os, ad);
		break;
	case 6:
		game_colliding_op6(os, ad);
		break;
	case 7:
		game_colliding_op7(os, ad);
		break;
	case 8:
		game_colliding_op8(os, ad);
		break;
	case 9:
		game_colliding_op9(os, ad);
		break;
	case 11:
		game_colliding_op11(os, ad);
		break;
	case 12:
		game_colliding_op12(os, ad);
		break;
	case 13:
		game_colliding_op13(os, ad);
		break;
	case 14:
		game_colliding_op14(os, ad);
		break;
	case 15:
		game_colliding_op15(os, ad);
		break;
	case 16:
		game_colliding_op16(os, ad);
		break;
	case 17:
		game_colliding_op17(os, ad);
		break;
	case 18:
		game_colliding_op18(os, ad);
		break;
	case 19:
		game_colliding_op19(os, ad);
		break;
	case 20:
		game_colliding_op20(os, ad);
		break;
	case 21:
		game_colliding_op21(os, ad);
		break;
	case 22:
		game_colliding_op22(os, ad);
		break;
	case 23:
		game_colliding_op23(os, ad);
		break;
	case 24:
		game_colliding_op0(os, ad);
		break;
	case 25:
		game_colliding_op25(os, ad);
		break;
	case 26:
		game_colliding_op26(os, ad);
		break;
	case 27:
		game_colliding_op27(os, ad);
		break;
	case 28:
		game_colliding_op28(os, ad);
		break;
	case 29:
		game_colliding_op29(os, ad);
		break;
	case 32:
		game_colliding_op32(os, ad);
		break;
	case 33:
		game_colliding_op33(os, ad);
		break;
	case 34:
		game_colliding_op34(os, ad);
		break;
	case 35:
		game_colliding_op35(os, ad);
		break;
	case 38:
		game_colliding_op38(os, ad);
		break;
	case 39:
		game_colliding_op39(os, ad);
		break;
	case 40:
		game_colliding_op40(os, ad);
		break;
	case 41:
		game_colliding_op41(os, ad);
		break;
	case 42:
		game_colliding_op42(os, ad);
		break;
	default:
		warning("game.collides_op colliding_opcode %d", ad->colliding_opcode);
		break;
	}
}

void game_collides_op_oscar(object_state_t *os, anim_data_t *ad) {
	switch (ad->colliding_opcode) {
	case 0:
		game_colliding2_op0(os, ad);
		break;
	case 1:
		game_colliding_op1(os, ad);
		break;
	case 2:
		game_colliding_op2(os, ad);
		break;
	case 3:
		game_colliding_op3(os, ad);
		break;
	case 4:
		game_colliding_op4(os, ad);
		break;
	case 5:
		game_colliding_op5(os, ad);
		break;
	case 6:
		game_colliding_op6(os, ad);
		break;
	case 7:
		game_colliding_op7(os, ad);
		break;
	case 8:
		game_colliding_op8(os, ad);
		break;
	case 9:
		game_colliding_op9(os, ad);
		break;
	case 14:
		game_colliding_op14(os, ad);
		break;
	case 15:
		game_colliding_op15(os, ad);
		break;
	case 16:
		game_colliding_op16(os, ad);
		break;
	case 17:
		game_colliding_op17(os, ad);
		break;
	case 20:
		game_colliding_op20(os, ad);
		break;
	case 22:
		game_colliding_op22(os, ad);
		break;
	case 23:
		game_colliding_op23(os, ad);
		break;
	case 28:
		game_colliding_op28(os, ad);
		break;
	case 32:
		game_colliding_op32(os, ad);
		break;
	case 33:
		game_colliding_op33(os, ad);
		break;
	case 34:
		game_colliding_op34(os, ad);
		break;
	case 35:
		game_colliding_op35(os, ad);
		break;
	default:
		warning("game.collides_op_oscar colliding_opcode %d", ad->colliding_opcode);
		break;
	}
}

