/*
 * Carlos engine rewrite
 * Copyright (C) 2006-2011 Gregory Montoir (cyx@users.sourceforge.net)
 */

#ifndef RENDER_H__
#define RENDER_H__

#include "util.h"

#define kMaxSpritesCount 128

struct sprite_t {
	const uint8 *data;
	int num;
	int x, y;
	int flags;
	int pal_num;
};

struct offscreen_buffer_t {
	uint8 *ptr;
	int size;
	int pitch;
};

struct Texture;
struct SpriteTexture;

struct Render {
	int _w, _h;
	int _scaler;
	uint16_t _tex8to5551[256];
	uint16_t *_texBuf;
	offscreen_buffer_t offscreen_buffers_table[2];
	int _tileMapOriginX, _tileMapOriginY;
	uint8 *_tilesBuffer;
	const uint8 *_tileMapAnimLookupTable, *_tileMapData;
	int _tileMapW, _tileMapH;
	sprite_t _spritesList[kMaxSpritesCount];
	int _spritesListCount;
	Texture *_spritesTex[kMaxSpritesCount];

	void init(int w, int h);
	void destroy();
	void flushTextures();
	void setPalette(const uint16 *pal_data, int num_colors, int index = 0);
	void updateScreen();
	void setTileData(const uint8 *tile_data, int transparentColor);
	void setTileMapData(const uint8 *map_data, int w, int h, const uint8 *anim_lut);
	void setTileMapOrigin(int x, int y);
	void setCopper(const uint16 *pal_data, int num_lines);
	void setBackgroundFrame(const uint8 *frame_data, int w, int h, const uint8 *lut);
	void clearSpriteList();
	void addToSpriteList(const uint8 *data, int num, int x, int y, int pal = 0);
	void convertTexture(uint16 *dst, int dstPitch, const uint8 *src, int srcPitch, int w, int h, const uint16 *lut);
	void uploadSpriteTexture(const uint8 *data, int pal, SpriteTexture *cacheTex);
	void drawSprite(const uint8 *spr_data, int x, int y, int w, int h, int flags, int pal_num, uint8 *dst, int dstPitch);
};

#endif // RENDER_H__
