/*
 * Carlos engine rewrite
 * Copyright (C) 2006-2011 Gregory Montoir (cyx@users.sourceforge.net)
 */

#include "fileio.h"
#include "game.h"
#include "globals.h"
#include "render.h"
#include "resource.h"

const uint8 *Resource::get_new_ptr(int num) {
	int offset = 0;
	for (int i = 0; i < num; ++i) {
		const int sz = _new_new[offset] * _new_new[offset + 1];
		offset += sz + 2;
	}
	return _new_new + offset;
}

anim_data_t *Resource::get_ref_ptr(int num_ref, int anim_count) {
	anim_data_t *ad = 0;
	if (num_ref < 0) {
		ad = &_ref_ref_use_ptr3;
	} else if (num_ref > 0) {
		assert(num_ref < anim_count);
		ad = &_ref_ref[num_ref];
	}
	return ad;
}

void Resource::get_sprite_dim(const uint8 *data, int num, int *w, int *h) {
	const int offs = READ_LE_UINT16(data + num * 2);
	*w = READ_LE_UINT16(data + offs);
	*h = READ_LE_UINT16(data + offs + 2);
}

const uint8 *Resource::fetch_level_data(const char *filename, int *size) {
	int slot_tmp = fio_open(filename, FIO_READ, 1);
	assert(res_slot_count < MAX_RES_SLOTS);
	res_slot_table[res_slot_count] = slot_tmp;
	++res_slot_count;
	const uint8 *data = fio_fetch(slot_tmp);
	if (size) {
		*size = fio_size(slot_tmp);
	}
	return data;
}

void Resource::load_ref_ref__v2(const uint8 *data, int size) {
	static const uint16 move_delta_offsets_table[15] = {
		0, 88, 92, 180, 236, 292, 374, 462, 518, 570, 682, 768, 854, 898, 942
	};
	const int anim_count = size / 0x40;
	assert((size % 0x40) == 0);
	assert(anim_count <= (int)ARRAYSIZE(_ref_ref));
	for (int i = 0; i < anim_count; ++i) {
		anim_data_t *ad = &_ref_ref[i];
		ad->unk0 = *data++;
		ad->lifes = *data++;
		ad->cycles = *data++;
		ad->unk3 = *data++;
		ad->unk4 = *data++;
		ad->unk5 = *data++;
		ad->unk6 = *data++;
		ad->init_sprite_num = *data++;
		ad->colliding_opcode = *data++;
		ad->logic_opcode = *data++;
		ad->sound_num = *data++;
		ad->rnd = *data++;
		ad->sprite_num = *data++;
		ad->sprite_flags = *data++;
		ad->anim_w = READ_LE_UINT16(data); data += 2;
		ad->anim_h = READ_LE_UINT16(data); data += 2;
		ad->score = READ_LE_UINT16(data); data += 2;
		ad->bounding_box_x1 = READ_LE_UINT16(data); data += 2;
		ad->bounding_box_x2 = READ_LE_UINT16(data); data += 2;
		ad->bounding_box_y1 = READ_LE_UINT16(data); data += 2;
		ad->bounding_box_y2 = READ_LE_UINT16(data); data += 2;
		const int index = READ_LE_UINT16(data); data += 2;
		ad->move_data_ptr = (index != 0) ? move_delta_table + move_delta_offsets_table[index - 1] : 0;
		data += 2;
		ad->distance_dx = READ_LE_UINT16(data); data += 2;
		ad->distance_dy = READ_LE_UINT16(data); data += 2;
		ad->anim_data1_ptr = get_ref_ptr((int16)READ_LE_UINT16(data), anim_count); data += 2;
		data += 2;
		ad->anim_data2_ptr = get_ref_ptr((int16)READ_LE_UINT16(data), anim_count); data += 2;
		data += 2;
		ad->anim_data3_ptr = get_ref_ptr((int16)READ_LE_UINT16(data), anim_count); data += 2;
		data += 2;
		ad->dx = READ_LE_UINT16(data); data += 2;
		ad->dy = READ_LE_UINT16(data); data += 2;
		ad->anim_data4_ptr = get_ref_ptr((int16)READ_LE_UINT16(data), anim_count); data += 2;
		data += 2;
		ad->dx2 = READ_LE_UINT16(data); data += 2;
		ad->dy2 = READ_LE_UINT16(data); data += 2;
		ad->draw_dx = READ_LE_UINT16(data); data += 2;
		ad->draw_dy = READ_LE_UINT16(data); data += 2;
	}
}

static const uint16 decor_pal_table[16 * 5] = {
	0x000, 0xFED, 0xCD1, 0x7B2, 0x470, 0x231, 0x000, 0xFC3,
	0x07B, 0xB30, 0x3BF, 0x26A, 0x136, 0xF94, 0x853, 0x511,
	0x000, 0xFED, 0xABB, 0x789, 0x556, 0x434, 0x000, 0xFC3,
	0x07B, 0xB30, 0x3BF, 0x26A, 0x136, 0xF94, 0xB64, 0x844,
	0x000, 0xFED, 0xABB, 0x789, 0x556, 0x434, 0x000, 0xFC3,
	0x07B, 0xB30, 0x3BF, 0x26A, 0x136, 0xF94, 0xB64, 0x844,
	0x000, 0xFEC, 0xF00, 0x7BB, 0x9C5, 0xC97, 0x634, 0xDD3,
	0x07B, 0xA75, 0x045, 0x079, 0x493, 0x655, 0x221, 0xEB9,
	0x000, 0xFFF, 0xF00, 0x78D, 0x8C3, 0x642, 0xFC9, 0xED0,
	0x07B, 0x878, 0x221, 0x056, 0x086, 0x420, 0xA92, 0xFF7
};

static const uint16 lifebar_pal[16] = {
	0x000, 0xFFF, 0x000, 0xF00, 0xF10, 0xF20, 0xF30, 0xF40,
	0xF50, 0xF60, 0xF70, 0xF80, 0xF90, 0xFA0, 0xFB0, 0xFC0
};

void Resource::set_colors(int level, Render *render) {
	/* set level palette */
	render->setPalette(&decor_pal_table[16 * level], 16);

	render->setPalette(lifebar_pal, 16, 1);

	/* background */
	switch (level) {
	case 0:
		render->setCopper(DEGRA00_COP, 160);
		break;
	case 1:
		render->setCopper(DEGRA01_COP, 160);
		break;
	case 2:
		render->setCopper(DEGRA03_COP, 160);
		break;
	case 3:
		render->setCopper(DEGRA04_COP, 160);
		break;
	case 4:
		render->setCopper(DEGRA50_COP, 160);
		break;
	}
}

void Resource::load_level_data__v3(int level, Render *render) {
	int slot_tmp, size;
	char filename[16];

	unload_level_data();

	/* tiles gfx */
	static const uint8 blk[5] = { 1, 2, 2, 4, 5 };
	snprintf(filename, sizeof(filename), "AA%dBLK.SQX", blk[level]);
	_decor_blk = fetch_level_data(filename);
	static const uint8 clr[5] = { 0, 8, 0, 8, 0 };
	render->setTileData(g_game._res._decor_blk, clr[level]);

	/* tiles flags */
	snprintf(filename, sizeof(filename), "AA%dREF.SQX", level + 1);
	_decor_ref = fetch_level_data(filename);

	/* tile map (will be modified, make a copy) */
	snprintf(filename, sizeof(filename), "A%d0CDG.SQX", level + 1);
	slot_tmp = fio_open(filename, FIO_READ, 1);
	const uint8 *file_data = fio_fetch(slot_tmp);
	memcpy(_decor_cdg, file_data, fio_size(slot_tmp));
	fio_close(slot_tmp);

	/* tiles update data */
	snprintf(filename, sizeof(filename), "A%d0NEW.SQX", level + 1);
	_new_new = fetch_level_data(filename);

	/* carlos sprites */
	static const uint8 s0x[] = { 0, 2, 2, 6, 8 };
	snprintf(filename, sizeof(filename), "S0%d.SQX", s0x[level]);
	_carlos1_spr = fetch_level_data(filename);
	snprintf(filename, sizeof(filename), "S0%d.SQX", s0x[level] + 1);
	_carlos2_spr = fetch_level_data(filename);

	/* monster/level sprites */
	snprintf(filename, sizeof(filename), "S%d.SQX", 11 + level);
	_level_spr = fetch_level_data(filename);

	/* boss sprites */
	snprintf(filename, sizeof(filename), "S%d.SQX", 21 + level);
	_boss_spr = fetch_level_data(filename);

	/* oscar sprites */
	snprintf(filename, sizeof(filename), "S%d.SQX", 31 + level);
	_oscar_spr = fetch_level_data(filename);

	/* friend sprites */
	snprintf(filename, sizeof(filename), "S%d.SQX", 41 + level);
	_friend_spr = fetch_level_data(filename);

	/* monster flags */
	snprintf(filename, sizeof(filename), "A%d0POS.SQX", level + 1);
	_posit_ref = fetch_level_data(filename, &size);
	assert((size % 10) == 0);
	_posit_count = size / 10;

	/* monster anim_data */
	snprintf(filename, sizeof(filename), "REF%dREF.SQX", level + 1);
	slot_tmp = fio_open(filename, FIO_READ, 1);
	file_data = fio_fetch(slot_tmp);
	load_ref_ref__v2(file_data, fio_size(slot_tmp));
	fio_close(slot_tmp);

	_screen_cdg_tile_map_h = READ_LE_UINT16(_posit_ref + 2);
	_screen_cdg_tile_map_w = READ_LE_UINT16(_posit_ref + 4);
	_screen_cdg_tile_map_start_y = READ_LE_UINT16(_posit_ref + 6);
	_screen_cdg_tile_map_start_x = READ_LE_UINT16(_posit_ref + 8);
	--_posit_count;
	_posit_ref += 10;
	if (READ_LE_UINT16(_posit_ref) == 0xFFFE) {
		_screen_cdg_tile_map_h2 = READ_LE_UINT16(_posit_ref + 2);
		const int w = READ_LE_UINT16(_posit_ref + 4);
		load_fond(w & 0x7F, render);
#if 0 // v3
		word_183D9 = w & 0x80;
#endif
		--_posit_count;
		_posit_ref += 10;
	}

	render->setTileMapData(_decor_cdg, _screen_cdg_tile_map_w, _screen_cdg_tile_map_h, _screen_cdg_offsets);

	level_data_loaded = 1;
}

void Resource::unload_level_data() {
	if (level_data_loaded) {
		for (int i = 0; i < res_slot_count; ++i) {
			fio_close(res_slot_table[i]);
		}
		memset(res_slot_table, 0, sizeof(res_slot_table));
		res_slot_count = 0;
		level_data_loaded = 0;
	}
}

void Resource::load_fond(int num, Render *render) {
	char filename[32];
	snprintf(filename, sizeof(filename), "FOND0%d.SQX", num);
	int slot = fio_open(filename, FIO_READ, 0);
	if (slot >= 0) {
		uint8 t[16];
		for (int i = 0; i < 16; ++i) {
			t[i] = i;
		}
		switch (num) {
		case 0:
			t[10] = t[11] = 8;
			break;
		case 3:
			t[1] = t[3] = 8;
			break;
		case 4:
			t[1] = t[3] = t[11] = t[13] = 8;
			break;
		}
		const uint8 *file_data = fio_fetch(slot);
		render->setBackgroundFrame(file_data, 640, 32, t);
		fio_close(slot);
	}
}

void Resource::init_static_data__v3() {
	memset(&anim_data0, 0, sizeof(anim_data0));
	anim_data0.unk0 = -1;
	anim_data0.sound_num = -1;
}

