/*
 * Carlos engine rewrite
 * Copyright (C) 2006-2011 Gregory Montoir (cyx@users.sourceforge.net)
 */

#include "util.h"

struct SqxDecoder {
	uint16 code;
	int carry;
	uint8 c1, c2, c3;
	const uint8 *src;
	uint8 *dst;
};

static void rcr(uint16 *reg, int *carry) {
	int c = (*reg & 1) == 1;
	*reg >>= 1;
	if (*carry) {
		*reg |= 0x8000;
	}
	*carry = c;
}

static void rcl(uint16 *reg, int *carry) {
	int c = (*reg & 0x8000) == 0x8000;
	*reg <<= 1;
	if (*carry) {
		*reg |= 0x1;
	}
	*carry = c;
}

static void shr(uint16 *reg, int *carry) {
	*carry = (*reg & 1) == 1;
	*reg >>= 1;
}

static void cpy(SqxDecoder *sqx, int offset, int len) {
	const uint8 *src = sqx->dst + offset;
	for (int i = 0; i < len + 2; ++i) {
		*sqx->dst++ = *src++;
	}
}

static int sqx_decode_helper(SqxDecoder *sqx, int type) {
	uint16 offset, len = 0;

	switch (type) {
	case 0:
		*sqx->dst++ = *sqx->src++;
		return 0;
	case 1:
		shr(&sqx->code, &sqx->carry);
		if (sqx->code == 0) {
			sqx->code = READ_LE_UINT16(sqx->src); sqx->src += 2;
			rcr(&sqx->code, &sqx->carry);
			rcl(&len, &sqx->carry);
			shr(&sqx->code, &sqx->carry);
		} else {
			rcl(&len, &sqx->carry);
			shr(&sqx->code, &sqx->carry);
			if (sqx->code == 0) {
				sqx->code = READ_LE_UINT16(sqx->src); sqx->src += 2;
				rcr(&sqx->code, &sqx->carry);
			}
		}
		rcl(&len, &sqx->carry);
		offset = 0xFF00 | *sqx->src++;
		cpy(sqx, (int16)offset, len);
		return 0;
	case 2:
		offset = READ_LE_UINT16(sqx->src); sqx->src += 2;
		len = offset & 0xFF;
		offset >>= sqx->c1;
		offset |= sqx->c2 << 8;
		len &= sqx->c3;
		if (len == 0) {
			len = *sqx->src++;
			if (len == 0) {
				return 1;
			}
		}
		cpy(sqx, (int16)offset, len);
		return 0;
	}
	return 1;
}

int sqx_decode(const uint8 *src, uint8 *dst) {
	uint8 j1 = *src++;
	uint8 j2 = *src++;
	uint8 j3 = *src++;
	assert(j1 + j2 + j3 == 3 && j1 != j2 && j2 != j3 && j3 != j1);
	SqxDecoder sqx;
	memset(&sqx, 0, sizeof(SqxDecoder));
	sqx.c1 = *src++;
	sqx.c2 = 0;
	sqx.c3 = 0;
	sqx.src = src;
	sqx.dst = dst;
	for (int i = 0; i < sqx.c1; ++i) {
		sqx.c2 >>= 1;
		sqx.c2 |= (1 << 7);
		sqx.c3 <<= 1;
		sqx.c3 |= 1;
	}
	sqx.code = 1;
	int end = 0;
	while (!end) {
		shr(&sqx.code, &sqx.carry);
		if (sqx.code == 0) {
			sqx.code = READ_LE_UINT16(sqx.src); sqx.src += 2;
			rcr(&sqx.code, &sqx.carry);
		}
		if (!sqx.carry) {
			end = sqx_decode_helper(&sqx, j1);
		} else {
			shr(&sqx.code, &sqx.carry);
			if (sqx.code == 0) {
				sqx.code = READ_LE_UINT16(sqx.src); sqx.src += 2;
				rcr(&sqx.code, &sqx.carry);
			}
			if (!sqx.carry) {
				end = sqx_decode_helper(&sqx, j2);
			} else {
				end = sqx_decode_helper(&sqx, j3);
			}
		}
	}
	return sqx.dst - dst;
}
